#include <macros.h>
#include <string.h>
#include <osbind.h>
#include <mintbind.h>
#include <signal.h>
#include <basepage.h>

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;

#include "scsicmd.h"
#include "filesys.h"
#include "atarierr.h"
#include "kmacros.h"
#include "devproto.h"

short scsidev=3;
int testunit=1;
int last_cache_entry=-1;

uchar dmabuf[CDROM_BLOCKSIZE];
short inittime,initdate;
struct kerinfo *kernel=NULL;
extern char *version;

#define CD_MAX_READ 256

DEVDRV cdrom_dev={
  cd_open,cd_write,cd_read,cd_lseek,
  cd_ioctl,cd_datime,cd_close,cd_select,cd_unselect
};

struct dev_descr devinfo = {
  &cdrom_dev,0,0,(struct tty *)0l,
  {0l,0l,0l,0l}
  };

int read_block(int device, long blockno, char *buf, long nblocks) {
  int retcode;
#ifndef NO_DEBUG
  extern char *sense_msg[];
#endif

  if(testunit)
    if((retcode=test_unit_ready(device)))
      goto jumphere;

again:
  TRACE(("read_block(%d, %ld, %lx, %lu)",device,blockno,buf,nblocks));
  if((nblocks==1)&&(blockno==last_cache_entry)) {
    if(buf)
      memcpy(buf,dmabuf,CDROM_BLOCKSIZE);
    return 0;
  }
  if(nblocks==1) {
    retcode=scsi_read(device,blockno,1,CDROM_BLOCKSIZE,dmabuf);
    DEBUG(("Returning from scsi_read"));
  } else {
    DEBUG(("read_block: %ld, nblocks %ld, buf %06lx",
	   blockno,nblocks,buf));
    retcode=scsi_read(device,blockno,nblocks,CDROM_BLOCKSIZE,buf);
    DEBUG(("Returning from scsi_read"));
  }
 jumphere:
  if(retcode) {
    DEBUG(("read_block: scsi_read returned %d",retcode));
    if((retcode&0x1e)==8) {
      DEBUG(("read_block: Device BUSY"));
      return EDRVNR;		/* Return Drive Not Ready */
    } else if((retcode&0x1e)==2) { /* OK, we will request sense... */
      SenseData sensedata;
      request_sense(device, &sensedata);
      if(sensedata.nonextendedsense.ErrorClass<7) {
	ALERT(("read_block: Nonextended sense: class %u, code %u",
	      sensedata.nonextendedsense.ErrorClass,
	      sensedata.nonextendedsense.ErrorCode));
	if(sensedata.nonextendedsense.Valid)
	  ALERT(("read_block: in block %lu",
		((long)sensedata.nonextendedsense.BlockMSB<<16)|
		sensedata.nonextendedsense.BlockLSB));
	return EDRVNR;
      } else if(!sensedata.extendedsense.ErrorCode) {
	DEBUG(("read_block: Request sense: %s",
	      sense_msg[sensedata.extendedsense.SenseKey]));
	if(sensedata.extendedsense.Valid)
	  DEBUG(("read_block: in block %lu",
		sensedata.extendedsense.InfoBytes));
	switch(sensedata.extendedsense.SenseKey) {
	case 0:			/* NO SENSE */
	  DEBUG(("read_block: Got \"CHECK CONDITION\" - but no sense key !"));
	  goto OK;		/* The command seems to have executed
				   nice anyway - strange ! */
	  break;
	case 1:			/* RECOVERED ERROR */
	  goto OK;
	  break;
	case 2:			/* DEVICE NOT READY */
	  return EDRVNR;
	  break;
	case 3:			/* MEDIUM ERROR */
	  return EREADF;
	  break;
	case 4:			/* HARDWARE ERROR */
	  return EREADF;
	  break;
	case 5:			/* ILLEGAL REQUEST */
	  return EBADRQ;
	  break;
	case 6:			/* MEDIA CHANGED */
	  return E_CHNG;
	  break;
	case 7:			/* DATA PROTECT */
	  return EACCDN;
	  break;
	case 8:			/* BLANK CHECK */
	  return EWRITF;
	  break;
	case 9:			/* Device specific */
	  return ERROR;		/* Maybe return E_CHNG ? */
	  break;
	case 10:		/* COPY ABORTED */
	  return ERROR;
	  break;
	case 11:		/* ABORTED COMMAND, try again */
	  goto again;
	  break;
	case 12:		/* EQUAL - hmmm. Lets return error anyway */
	  return ERROR;
	  break;
	case 13:		/* VOLUME OVERFLOW */
	  return EWRITF;
	  break;
	case 14:		/* MISCOMPARE */
	  return ERROR;
	  break;
	case 15:		/* Reserved. */
	  return ERROR;
	  break;
	}
      }
    }
    return ERROR;		/* Probably a timeout */
  }
 OK:
  if(nblocks==1) {
    if(buf) {
      memcpy(buf,dmabuf,CDROM_BLOCKSIZE);
    }
    last_cache_entry=blockno;
  }
  return 0;
}

long cd_open(FILEPTR *f) {
  TRACE(("cd_open(%lx)",f));
  f->pos=0;
  return 0;
}

long cd_write(FILEPTR *f, const char *buf, long bytes) {
  return EWRPRO;
}

long cd_read(FILEPTR *f, char *buf, long bytes) {
  long error;
  long block;
  long nblocks;
  long offset=f->pos%CDROM_BLOCKSIZE;
  long readchunk;
  long readsofar=0;

  DEBUG(("cd_read(FILE %lx,BUF %lx,BYTES %ld, offset %ld)",
	 f,buf,bytes,f->pos));

  block=f->pos/CDROM_BLOCKSIZE;  

  if(bytes<=0)
    return 0l;

  if(offset) {
    readchunk=min(CDROM_BLOCKSIZE-offset,bytes);
    error=read_block(scsidev, block, NULL,1);
    DEBUG(("cd_read: %ld bytes from end of block %ld offset %ld",
	   readchunk,block,offset));
    if(error)
      return 0;
    memcpy(buf,(char *)(dmabuf+offset),readchunk);
    bytes-=readchunk;
    buf+=readchunk;
    block++;
    readsofar=readchunk;
    f->pos+=readchunk;
  }
  nblocks=bytes/CDROM_BLOCKSIZE;
  if(nblocks) {
    while(nblocks>CD_MAX_READ) {
      readchunk=(long)CD_MAX_READ*(long)CDROM_BLOCKSIZE;
      TRACE(("cd_read: Reading %d blocks, $%lx bytes.",
	     CD_MAX_READ,readchunk));
      error=read_block(scsidev,block,buf,CD_MAX_READ);
      if(error)
	return readsofar;
      buf+=readchunk;
      bytes-=readchunk;
      block+=CD_MAX_READ;
      nblocks-=CD_MAX_READ;
      readsofar+=readchunk;
      f->pos+=readchunk;
    }
    if(!nblocks)
      goto readdone;
    DEBUG(("cd_read: Last multiple %ld blocks from block %ld to buf %06lx",
	   nblocks,block,buf));
    error=read_block(scsidev,block,buf,(unsigned)nblocks);
    if(error)
      return readsofar;
    readchunk=nblocks*((long)CDROM_BLOCKSIZE);
    buf=(char *)(buf+readchunk);
    bytes-=readchunk;
    block+=nblocks;
    readsofar+=readchunk;
    f->pos+=readchunk;
    nblocks=0;
  }
  if(bytes) {
    error=read_block(scsidev,block,NULL,1);
    DEBUG(("cd_read : Reading %ld bytes from start of last block %ld",
	   bytes,block));
    if(error)
      return readsofar;
    memcpy(buf,dmabuf,bytes);
    readsofar+=bytes;
    f->pos+=bytes;
    bytes=0;
  }
 readdone:
  return readsofar;
}

long cd_lseek(FILEPTR *f, long where, int whence) {
  TRACE(("cd_lseek(%lx,%ld,%d)",f,where,whence));
				/* OBS: Does not check for CD-SIZE !!! */
  switch(whence) {
  case SEEK_SET:
    if(where>=0)
      f->pos=where;
    else
      return ERANGE;
    break;
  case SEEK_CUR:
    if(f->pos+where>=0)
      f->pos+=where;
    else
      return ERANGE;
    break;
  case SEEK_END:
    return EINVFN;
    break;
  default:
    return EINVFN;
  }
  return f->pos;
}

long cd_ioctl(FILEPTR *f, int mode, void *buf) {
  TRACE(("cd_ioctl(%lx,%d,%lx)",f,mode,buf));
  switch(mode) {
  case FIONREAD:
    *((long *)buf)=(long)CD_MAX_READ*(long)CDROM_BLOCKSIZE;
    return 0;
  case FIONWRITE:
    *((long *)buf)=0l;
    return 0;
  default:
    return EINVFN;
  }
}

long cd_datime(FILEPTR *f, short *timeptr, int rwflag) {
  TRACE(("cd_datime(%lx,%lx,%d)",f,timeptr,rwflag));
  if(rwflag)
    return EWRPRO;

  timeptr[0]=inittime;
  timeptr[1]=initdate;
  return 0l;
}

long cd_close(FILEPTR *f, int pid) {
  return 0l;
}

long cd_select(FILEPTR *f, long proc, int mode) {
				/* Really... */
  switch(mode) {
  case O_WRONLY:
    return 1;			/* Might as well let it have the error */
    break;			/* code at once... */
  case O_RDONLY:
    return 1;
    break;
  }
  TRACE(("cd_select()"));
  return 1;
}

void cd_unselect(FILEPTR *f, long proc, int mode) {
  TRACE(("cd_unselect()"));
  return;
}

DEVDRV *cddev_init(struct kerinfo *k) {
  kernel=k;
  CCONWS("\033pCD-ROM device driver\033q\r\nVersion ");
  CCONWS(version);
  CCONWS("\r\n(c) Erling Henanger 1994\r\n");
  if(DCNTL(DEV_INSTALL,"u:\\dev\\cdrom",&devinfo)<=0l) {
    CCONWS("Unable to install cdrom device\r\n");
    return (DEVDRV *)0l;
  }
  inittime=TGETTIME();
  initdate=TGETDATE();
  CCONWS("CD-rom device installed\r\n");
  return (DEVDRV *)1l;
}
