#include <string.h>

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;
extern long status_timeout;
extern char *version;
extern long poll_timeout;

union clong {
	char	aschar[4];
	long	aslong;
};

typedef struct cookie {
	union clong tag;
	long value;
} COOKIE;
#define CJAR	((COOKIE **) 0x5a0L)

#define TT		0x00020000
#define FALCON		0x00030000	/* Changed to reflect ROMs LT */

#include "filesys.h"
#include "atarierr.h"
#include "kmacros.h"
#include "steal.h"
#include "sioctl.h"
#include "pollproto.h"

short inittime,initdate;
struct kerinfo *kernel;
extern short sleeptime;

#define SCSI_SIMPLE 0
#define SCSI_READ   0
#define SCSI_WRITE  1

extern long do_scsi(short device, short ncmd,
			   long nbytes, short read,
			   char *cmdbuf, char *dmabuf);

DEVDRV scsi_dev={
  scsi_open,scsi_write,scsi_read,scsi_lseek,
  scsi_ioctl,scsi_datime,scsi_close,scsi_select,scsi_unselect
};

struct dev_descr devinfo = {
  &scsi_dev,0,0,(struct tty *)0l,
  {0l,0l,0l,0l}
  };

long do_simple_command(SCSI_COMMAND *cmd) {
  long old_status_timeout=status_timeout;
  status_timeout=cmd->timeout;
  cmd->cmdblk[1]|=((cmd->device&0x0700)>>3);
  DEBUG(("scsi simple: Timeout %ld, command: %02x %02x %02x %02x %02x %02x",
	 cmd->timeout,
	 cmd->cmdblk[0],cmd->cmdblk[1],cmd->cmdblk[2],
	 cmd->cmdblk[3],cmd->cmdblk[4],cmd->cmdblk[5]));
  cmd->status=(short)do_scsi(cmd->device&0x7,cmd->cmdlen,
				    0l,SCSI_SIMPLE, cmd->cmdblk,(char *)0l);
#ifndef NO_DEBUG
  if(cmd->status)
    DEBUG(("scsi simple returing %d",cmd->status));
#endif
  status_timeout=old_status_timeout;
  return 0l;
}

long do_reading_command(SCSI_COMMAND *cmd) {
  cmd->cmdblk[1]|=((cmd->device&0x0700)>>3);
  DEBUG(("scsi read: Timeout %ld, %02x %02x %02x %02x %02x %02x: length %ld, buf %lx",
	 cmd->timeout,
	 cmd->cmdblk[0],cmd->cmdblk[1],cmd->cmdblk[2],
	 cmd->cmdblk[3],cmd->cmdblk[4],cmd->cmdblk[5],
	 cmd->transferlen,cmd->buf));
  poll_timeout=cmd->timeout;
  cmd->status=do_scsi(cmd->device&0x7,cmd->cmdlen,
			     cmd->transferlen,SCSI_READ, cmd->cmdblk,cmd->buf);
#ifndef NO_DEBUG
  if(cmd->status)
    DEBUG(("scsi read returing %d",cmd->status));
#endif
  return 0l;
}

long do_writing_command(SCSI_COMMAND *cmd) {
  cmd->cmdblk[1]|=((cmd->device&0x0700)>>3);
  poll_timeout=cmd->timeout;
  DEBUG(("scsi write: Timeout %ld, %02x %02x %02x %02x %02x %02x",
	 cmd->timeout,
	 cmd->cmdblk[0],cmd->cmdblk[1],cmd->cmdblk[2],
	 cmd->cmdblk[3],cmd->cmdblk[4],cmd->cmdblk[5]));
  cmd->status=do_scsi(cmd->device&0x7,cmd->cmdlen,
		       cmd->transferlen,SCSI_WRITE, cmd->cmdblk,cmd->buf);
#ifndef NO_DEBUG
  if(cmd->status)
    DEBUG(("scsi write returing %d",cmd->status));
#endif
  return 0l;
}

long scsi_open(FILEPTR *f) {
  TRACE(("scsi_open(%lx)",f));
  f->pos=0;
  return 0;
}

long scsi_write(FILEPTR *f, const char *buf, long bytes) {
  return EINVFN;
}

long scsi_read(FILEPTR *f, char *buf, long bytes) {
  return EINVFN;
}

long scsi_lseek(FILEPTR *f, long where, int whence) {
  return EINVFN;
}

long scsi_ioctl(FILEPTR *f, int mode, void *buf) {
  TRACE(("scsi_ioctl(%lx,%d,%lx)",f,mode,buf));
  switch(mode) {
  case SSCMD:
    return do_simple_command((SCSI_COMMAND *)buf);
    break;
  case SRCMD:
    return do_reading_command((SCSI_COMMAND *)buf);
    break;
  case SWCMD:
    return do_writing_command((SCSI_COMMAND *)buf);
    break;
  case SSLEEP:
    sleeptime=(short)*(long *)buf;
    TRACE(("scsi_ioctl: Sleeptime %d",sleeptime));
    return 0l;
    break;
  case FIONREAD:
    *((long *)buf)=0l;
    return 0l;
  case FIONWRITE:
    *((long *)buf)=0l;
    return 0l;
  default:
    return EINVFN;
  }
}

long scsi_datime(FILEPTR *f, short *timeptr, int rwflag) {
  TRACE(("scsi_datime(%lx,%lx,%d)",f,timeptr,rwflag));
  if(rwflag)
    return EWRPRO;

  timeptr[0]=inittime;
  timeptr[1]=initdate;
  return 0l;
}

long scsi_close(FILEPTR *f, int pid) {
  return 0l;
}

long scsi_select(FILEPTR *f, long proc, int mode) {
  TRACE(("scsi_select()"));
  switch(mode) {
  case O_WRONLY:
    return 1;
  case O_RDONLY:
    return 1;
  }
  DEBUG(("scsi_select: illegal mode"));
  return 1;
}

void scsi_unselect(FILEPTR *f, long proc, int mode) {
  return;
}

DEVDRV *scsi_init(struct kerinfo *k) {
  char *devname="u:\\dev\\scsi";
  COOKIE *jar;
  short machine=1;
  kernel=k;
  jar = *CJAR;
  if (jar) {
    while (jar->tag.aslong != 0) {
      if (!strncmp(jar->tag.aschar, "_MCH",4)) {
	if (jar->value == FALCON) {
	  machine = 1;
	} else if (jar->value == TT) {
	  machine = 0;
	} else {
	  machine = -1;
	}
	break;
      }
      jar++;
    }
  }
  CCONWS("\033pPolling SCSI device driver\033q\r\n");
  if(machine==1) {
    CCONWS("Falcon ");
  } else if (!machine) {
    CCONWS("TT ");
  } else {
    CCONWS(" ");
  }
  CCONWS(version);
  CCONWS("\r\n(c) Erling Henanger 1994\r\n");
  if(machine&&machine!=1) {
    CCONWS("This SCSI device driver only works on Falcon and TT\r\n");
    CCONWS("Unable to install scsi device\r\n");
  }
  if(DCNTL(DEV_INSTALL,devname,&devinfo)<=0l) {
    CCONWS("Unable to install scsi device\r\n");
    return (DEVDRV *)0l;
  }
  inittime=TGETTIME();
  initdate=TGETDATE();
  CCONWS("SCSI device installed\r\n");
  return (DEVDRV *)1l;
}
