#include <stdio.h>
#include <string.h>
#include "sioctl.h"
#include "scsicmd.h"

typedef unsigned char  uchar;
typedef unsigned short ushort;
typedef unsigned long  ulong;

#include "filesys.h"
#ifndef NOKERNEL
#include "kmacros.h"
#endif
#include "atarierr.h"
#include "sioctl.h"

#ifndef NOKERNEL
extern struct kerinfo *kernel;
#else
#define NO_DEBUG
#include <mintbind.h>
#define FOPEN   Fopen
#define FCLOSE  Fclose
#define FCNTL   Fcntl
#define FSELECT Fselect
#ifndef NO_DEBUG
#define DEBUG(x) printf x
#else
#define DEBUG(x)
#endif
#define TRACE(x)
#define ALERT(x)
#endif /* NOKERNEL */

char minibuf[512];
union {
  unsigned long l;
  unsigned char c[4];
} foo;

SCSI_COMMAND the_cmd;
char cmdblk[12];
char *scsidev_file="u:\\dev\\scsi";

#define clrcmdblk() bzero(&cmdblk,12l)

char *our_error[]={
  "No error",
  "Timeout in select phase",
  "SCSI bus still busy",
  "Timeout polling \"DMA-done\"",
  "Timeout during acknowledge",
  "Timeout waiting for request",
  "Timeout writing databyte",
  "Timeout in command phase",
  "Flock semaphore in use",
  ""
  };
    

void explain_error(int device,int error) {
  SenseData    Sensedata;
  if((error&0x1e)==8) {
    printf("Device busy\n");
    return;
  }
  if((error&0x1e)==2) {
    request_sense(device,&Sensedata);
    printf("Valid: %d, FM: %d, EOM: %d, ILI: %d, ErrClass %d, ErrCode %d\n",
	   Sensedata.extendedsense.Valid,
	   Sensedata.extendedsense.FM,
	   Sensedata.extendedsense.EOM,
	   Sensedata.extendedsense.ILI,
	   Sensedata.extendedsense.ErrorClass,
	   Sensedata.extendedsense.ErrorCode);
    printf("SegmentNumber %d, InfoBytes %lx, ExtraBytes %d\n",
	   Sensedata.extendedsense.SegmentNumber,
	   Sensedata.extendedsense.InfoBytes,
	   Sensedata.extendedsense.ExtraBytes);
    printf("CommandSpec: %lx, additionalSense %d, additionalQualifier %d\n",
	   Sensedata.extendedsense.CommandSpec,
	   Sensedata.extendedsense.additionalSense,
	   Sensedata.extendedsense.additionalQualifier);
    printf("FieldReplacableUnitCode %d, SKSV %d, SenseSpecific %lx\n",
	   Sensedata.extendedsense.FieldReplaceableUnitCode,
	   Sensedata.extendedsense.SKSV,
	   Sensedata.extendedsense.SenseSpecific);
    if(Sensedata.nonextendedsense.ErrorClass<7) {
      printf("Nonextended sense: class %u, code %u",
	     Sensedata.nonextendedsense.ErrorClass,
	     Sensedata.nonextendedsense.ErrorCode);
      if(Sensedata.nonextendedsense.Valid) {
	printf(" in block %lu\n",
	       ((long)Sensedata.nonextendedsense.BlockMSB<<16)|
	       Sensedata.nonextendedsense.BlockLSB);
      } else {
	printf("\n");
      }
      return;
    }
    if(!Sensedata.extendedsense.ErrorCode) {
      printf(sense_msg[Sensedata.extendedsense.SenseKey]);
      if(Sensedata.extendedsense.Valid) {
	printf(" in block %lu\n",
	       Sensedata.extendedsense.InfoBytes);
      } else
	printf("\n");
      return;
    }
    printf("Unknown sense code\n");
  }
  if(error<0) {
    printf("Error: %s\n",our_error[-error]);
    return;
  }
  printf("Unknown error %d\n",error);
}

int set_sleeptime(long sleeptime) {
  long errcode;
  long arg=sleeptime;
  int fd;
  fd=FOPEN(scsidev_file,0);
  if(fd<0) {
    DEBUG(("Could not open %s",scsidev_file));
    return 1;
  }
  errcode=FCNTL(fd,(long)&arg, SSLEEP);
  FCLOSE(fd);
  if(errcode)
    return 1;
  return 0;
}

int scsi_cmd(SCSI_COMMAND *cmd, int rw) {
  long errcode;
  long selectmask;
  int fd;
  char *oldbuf=NULL;
  if(cmd->transferlen&&(cmd->transferlen<512l)&&!rw) {
    oldbuf=cmd->buf;
    cmd->buf=minibuf;
  }
  fd=FOPEN(scsidev_file,0);
  if(fd<0) {
    DEBUG(("Could not open %s",scsidev_file));
    return fd;
  }
  selectmask=(1<<fd);
  FSELECT(0,&selectmask,(long *)0l,0l);
  TRACE(("scsi: device %d unit %d cmdlen %d buflen %ld",
	 cmd->device&0xf,(cmd->device&0xf00)>>8,cmd->cmdlen,cmd->transferlen));
  TRACE(("commandblock: %02x %02x %02x %02x %02x %02x",
	 cmd->cmdblk[0],cmd->cmdblk[1],cmd->cmdblk[2],
	 cmd->cmdblk[3],cmd->cmdblk[4],cmd->cmdblk[5]));
  errcode=FCNTL(fd,cmd,(rw)?SWCMD:(cmd->transferlen)?SRCMD:SSCMD);
  FCLOSE(fd);
  if((!errcode)&&cmd->transferlen&&(cmd->transferlen<512l)&&!rw) {
    /* If minibuf used, copy interesting contents */
    memcpy(oldbuf,minibuf,cmd->transferlen);
    cmd->buf=oldbuf;
  }
  if(errcode)
    return (int)errcode;
  return the_cmd.status;
}

union {
  char asChar[6];
  struct {
    unsigned char opcode;
    unsigned LUN:3;
    unsigned reserved:4;
    unsigned Immed:1;
    unsigned short reserved2;
    unsigned reserved3:5;
    unsigned EOT:1;
    unsigned Retension:1;
    unsigned doLoad:1;
    unsigned char Control;
  } Loadpar;
} Loadcmd;

int load_unload(int device,int load, int immed, int eot, int retension,
		long timeout) {
  bzero(Loadcmd.asChar,6l);
  Loadcmd.Loadpar.opcode=0x1b;
  Loadcmd.Loadpar.LUN=(device>>8)&7;
  Loadcmd.Loadpar.Immed=(immed>0);
  Loadcmd.Loadpar.EOT=(eot>0);
  Loadcmd.Loadpar.Retension=(retension>0);
  Loadcmd.Loadpar.doLoad=load;
  memcpy(the_cmd.cmdblk,Loadcmd.asChar,6l);
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  the_cmd.transferlen=0l;
  the_cmd.buf=NULL;
  the_cmd.timeout=timeout;
  TRACE(("Load: SCSI Command %02x %02x %02x %02x %02x %02x\n",
	 Loadcmd.asChar[0],Loadcmd.asChar[1],Loadcmd.asChar[2],
	 Loadcmd.asChar[3],Loadcmd.asChar[4],Loadcmd.asChar[5]));
  return scsi_cmd(&the_cmd,0);
}

int scsi_seek(int device, long block,long timeout) {
  bzero(the_cmd.cmdblk,6l);
  foo.l=block;
  the_cmd.cmdblk[0]=CDSEEK;
  the_cmd.cmdblk[1]=foo.c[1]&0x1f;
  the_cmd.cmdblk[2]=foo.c[2];
  the_cmd.cmdblk[3]=foo.c[3];
  the_cmd.timeout=timeout;
  the_cmd.cmdlen=6;
  the_cmd.transferlen=0l;
  the_cmd.device=device;
  the_cmd.buf=(char *)0l;
  return scsi_cmd(&the_cmd,0);
}

union {
  uchar asChar[10];
  struct {
    uchar opcode;
    unsigned LUN:3;
    unsigned Reserved:2;
    unsigned BT:1;
    unsigned CP:1;
    unsigned Immed:1;
    uchar    Reserved2;
    ulong    block;
    uchar    Reserved3;
    uchar    Partition;
    uchar    Control;
  } Locate;
} Locatecmd;
  
int scsi_locate(int device, long block, int partition, long timeout) {
  bzero(the_cmd.cmdblk,10l);
  Locatecmd.Locate.opcode=SCSI_LOCATE;
  Locatecmd.Locate.LUN=(device>>8)&7;
  Locatecmd.Locate.Immed=0;
  Locatecmd.Locate.block=block;
  Locatecmd.Locate.Partition=partition;
  memcpy(the_cmd.cmdblk,Locatecmd.asChar,10l);
  the_cmd.timeout=timeout;
  the_cmd.cmdlen=10;
  the_cmd.transferlen=0l;
  the_cmd.device=device;
  the_cmd.buf=(char *)0l;
  return scsi_cmd(&the_cmd,0);
}  

union {
  uchar asChar[6];
  struct {
    uchar opcode;
    unsigned LUN:3;
    unsigned Reserved:3;
    unsigned SILI:1;
    unsigned Fixed:1;
    unsigned long Length:24;
    uchar    Control;
  } RWopt;
} SEQposition;

int scsi_seqread(short unit, long nblocks, long blocksize,
		 char *buf, long timeout) {
  bzero(SEQposition.asChar,6l);
  SEQposition.RWopt.opcode=8;
  SEQposition.RWopt.LUN=(unit>>8)&7;
  SEQposition.RWopt.SILI=0;
  SEQposition.RWopt.Fixed=1;
  SEQposition.RWopt.Length=nblocks;
  SEQposition.RWopt.Control=0;
  memcpy(the_cmd.cmdblk,SEQposition.asChar,6l);
  the_cmd.device=unit;
  the_cmd.timeout=timeout;
  the_cmd.transferlen=nblocks*blocksize;
  the_cmd.buf=buf;
  the_cmd.cmdlen=6;
  return scsi_cmd(&the_cmd,0);
}

int scsi_seqreadrew(short unit, long nblocks, long blocksize,
		 char *buf, long timeout) {
  bzero(SEQposition.asChar,6l);
  SEQposition.RWopt.opcode=0xf;
  SEQposition.RWopt.LUN=(unit>>8)&7;
  SEQposition.RWopt.SILI=0;
  SEQposition.RWopt.Fixed=1;
  SEQposition.RWopt.Length=nblocks;
  SEQposition.RWopt.Control=0;
  memcpy(the_cmd.cmdblk,SEQposition.asChar,6l);
  the_cmd.device=unit;
  the_cmd.timeout=timeout;
  the_cmd.transferlen=nblocks*blocksize;
  the_cmd.buf=buf;
  the_cmd.cmdlen=6;
  return scsi_cmd(&the_cmd,0);	/* NOTE: The bytes in this block */
				/* will ALSO be reversed !!! */
}

int scsi_seqwrite(short unit, long nblocks, long blocksize,
		 char *buf, long timeout) {
  bzero(SEQposition.asChar,6l);
  SEQposition.RWopt.opcode=0xa;
  SEQposition.RWopt.LUN=(unit>>8)&7;
  SEQposition.RWopt.SILI=0;
  SEQposition.RWopt.Fixed=1;
  SEQposition.RWopt.Length=nblocks;
  SEQposition.RWopt.Control=0;
  memcpy(the_cmd.cmdblk,SEQposition.asChar,6l);
  the_cmd.device=unit;
  the_cmd.timeout=timeout;
  the_cmd.transferlen=nblocks*blocksize;
  the_cmd.buf=buf;
  the_cmd.cmdlen=6;
  return scsi_cmd(&the_cmd,1);
}

union {
  uchar asChar[6];
  struct {
    uchar opcode;
    unsigned LUN:3;
    unsigned Reserved:3;
    unsigned WSmk:1;
    unsigned Immed:1;
    unsigned long Count:24;
    uchar Control;
  } markOpt;
} markCommand;

int scsi_filemark(int device, long count, long timeout) {
  bzero(markCommand.asChar,6l);
  markCommand.markOpt.opcode=0x10;
  markCommand.markOpt.LUN=(device>>8)&7;
  markCommand.markOpt.WSmk=0;
  markCommand.markOpt.Immed=0;
  markCommand.markOpt.Count=count;
  markCommand.markOpt.Control=0;
  memcpy(the_cmd.cmdblk,markCommand.asChar,6l);
  the_cmd.device=device;
  the_cmd.timeout=timeout;
  the_cmd.transferlen=0l;
  the_cmd.buf=(char *)0l;
  the_cmd.cmdlen=6;
  return scsi_cmd(&the_cmd,0);
}

union {
  uchar asChar[6];
  struct {
    uchar opcode;
    unsigned LUN:3;
    unsigned Reserved:4;
    unsigned Immed:1;
    uchar    Reserved2[3];
    uchar    Control;
  } Revopt;
} Rewcmd;

int scsi_rewind(short unit, int immed, long timeout) {
  bzero(Rewcmd.asChar,6l);
  Rewcmd.Revopt.opcode=1;
  Rewcmd.Revopt.LUN=(unit>>8)&7;
  Rewcmd.Revopt.Immed=(immed!=0);
  memcpy(the_cmd.cmdblk,Rewcmd.asChar,6l);
  the_cmd.device=unit;
  the_cmd.timeout=timeout;
  the_cmd.transferlen=0l;
  the_cmd.buf=(char *)0l;
  the_cmd.cmdlen=6;
  return scsi_cmd(&the_cmd,0);
}

union {
  uchar asChar[6];
  struct {
    uchar opcode;
    unsigned LUN:3;
    unsigned Reserved:2;
    unsigned Code:3;
    signed long Count:24;
    uchar Control;
  } spaceOpt;
} spaceCommand;

int scsi_space(int device, long count, short type, long timeout) {
  bzero(spaceCommand.asChar,6l);
  spaceCommand.spaceOpt.opcode=0x11;
  spaceCommand.spaceOpt.LUN=(device>>8)&7;
  spaceCommand.spaceOpt.Code=type;
  spaceCommand.spaceOpt.Count=count;
  DEBUG(("Spacing type %d, amount %ld - %02x%02x%02x\n",type,
	 spaceCommand.spaceOpt.Count,
	 spaceCommand.asChar[2],
	 spaceCommand.asChar[3],
	 spaceCommand.asChar[4]));
  memcpy(the_cmd.cmdblk,spaceCommand.asChar,6l);
  the_cmd.device=device;
  the_cmd.timeout=timeout;
  the_cmd.transferlen=0l;
  the_cmd.buf=(char *)0l;
  the_cmd.cmdlen=6;
  return scsi_cmd(&the_cmd,0);
}

union {
  uchar asChar[6];
  struct {
    uchar opcode;
    unsigned LUN:3;
    unsigned Reserved:3;
    unsigned Immed:1;
    unsigned Long:1;
    uchar    Reserved2[3];
    uchar    Control;
  } eraseOpt;
} eraseCmd;

int scsi_erase(short unit, int immed, int longflag, long timeout) {
  bzero(eraseCmd.asChar,6l);
  eraseCmd.eraseOpt.opcode=0x19;
  eraseCmd.eraseOpt.LUN=(unit>>8)&7;
  eraseCmd.eraseOpt.Immed=immed;
  eraseCmd.eraseOpt.Long=longflag;
  memcpy(the_cmd.cmdblk,eraseCmd.asChar,6l);
  the_cmd.device=unit;
  the_cmd.timeout=timeout;
  the_cmd.transferlen=0l;
  the_cmd.buf=(char *)0l;
  the_cmd.cmdlen=6;
  return scsi_cmd(&the_cmd,0);
}

int request_sense(int device,SenseData *buf) {
  bzero(the_cmd.cmdblk,12l);
  the_cmd.cmdblk[0]=CDREQSENSE;
  the_cmd.cmdblk[4]=(unsigned char)sizeof(SenseData);		/* allocation length */
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  the_cmd.transferlen=sizeof(SenseData);
  the_cmd.buf=(char *)buf;
  the_cmd.timeout=1000l;	/* 5 sec timeout should do... */
  TRACE(("request_sense: device %d",device));
  return scsi_cmd(&the_cmd,0);
}

int inquire(short device, InquireData *inq) {
  bzero(the_cmd.cmdblk,12l);
  the_cmd.cmdblk[0]=CDINQUIRE;
  the_cmd.cmdblk[4]=sizeof(InquireData);
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  the_cmd.transferlen=sizeof(InquireData);
  the_cmd.buf=(char *)inq;
  the_cmd.timeout=1000l;	/* 5 sec timeout should do for inquire as well */
  TRACE(("scsi_inquire: device %d",device));
  return scsi_cmd(&the_cmd,0);
}

int test_unit_ready(short device) {
  bzero(the_cmd.cmdblk,12l);
  the_cmd.cmdblk[0]=CDTSTUNIT;
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  the_cmd.transferlen=0l;
  the_cmd.buf=NULL;
  the_cmd.timeout=401l;
  TRACE(("test_unit_ready: device %d",device));
  return scsi_cmd(&the_cmd,0);
}

int rezero_unit(short device, long timeout) {
  bzero(the_cmd.cmdblk,12l);
  the_cmd.cmdblk[0]=ZEROUNIT;
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  the_cmd.transferlen=0l;
  the_cmd.buf=NULL;
  the_cmd.timeout=timeout;
  TRACE(("rezero_unit: device %d, timeout %ld",device,timeout));
  return scsi_cmd(&the_cmd,0);
}

int start_stop_unit(short device,short start,short immed,long timeout) {
  bzero(the_cmd.cmdblk,12l);
  the_cmd.cmdblk[0]=STARTSTOP;
  the_cmd.cmdblk[1]=immed;
  the_cmd.cmdblk[4]=(start)?1:0;
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  the_cmd.transferlen=0l;
  the_cmd.buf=NULL;
  the_cmd.timeout=timeout;
  TRACE(("start_stop_unit: device %d, start %d, immed %d, timeout %ld",
	 device,start,immed,timeout));
  return scsi_cmd(&the_cmd,0);
}

int prevent_allow_removal(short device,short prevent) {
  bzero(the_cmd.cmdblk,12l);
  the_cmd.cmdblk[0]=0x1e;
  the_cmd.cmdblk[4]=(prevent)?1:0;
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  the_cmd.transferlen=0l;
  the_cmd.buf=NULL;
  the_cmd.timeout=200;		/* 1 sec. timeout */
  TRACE(("prevent_allow_medium_removal: device %d, prevent %d",
	 device,prevent));
  return scsi_cmd(&the_cmd,0);
}

typedef union {
  uchar asChar[6];
  struct {
    uchar Opcode;
    ulong LUN:3;
    ulong Address:21;
    uchar Length;
    uchar Control;
  } Par;
} rwpar;

int scsi_read(short device, unsigned long startblock, unsigned nblocks,
	      unsigned long blocksize, char *buf) {
  int retcode=0;
  long readchunk;
  rwpar par;
  bzero(par.asChar,sizeof(par));
  par.Par.Opcode=CDREAD;
  if(!nblocks)
    nblocks=256;
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  while(nblocks>=256) {
    par.Par.Address=startblock;
    par.Par.Length=0;
    readchunk=256l*blocksize;
    memcpy(the_cmd.cmdblk,par.asChar,6l);
    the_cmd.transferlen=readchunk;
    the_cmd.buf=buf;
    the_cmd.timeout=30000l;
    retcode=scsi_cmd(&the_cmd,0);
    if(retcode)
      return retcode;
    buf+=readchunk;
    nblocks-=256;
    startblock+=256l;
  }
  if(nblocks) {
    par.Par.Address=startblock;
    par.Par.Length=(unsigned char)(nblocks&0xff);
    memcpy(the_cmd.cmdblk,par.asChar,sizeof(par));
    the_cmd.transferlen=(long)(nblocks*blocksize);
    the_cmd.buf=buf;
    the_cmd.timeout=30000l;
    retcode=scsi_cmd(&the_cmd,0);
  }
  return retcode;
}

int scsi_write(short device, long startblock, unsigned nblocks,
	       unsigned long blocksize, char *buf) {
  int retcode=0;
  long writechunk;
  rwpar par;
  bzero(par.asChar,sizeof(par));

  par.Par.Opcode=0x0a;		/* Write */
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  if(!nblocks)
    nblocks=256;
  while(nblocks>=256) {
    par.Par.Address=startblock;
    par.Par.Length=0;
    memcpy(the_cmd.cmdblk,par.asChar,sizeof(par));
    writechunk=256l*blocksize;
    the_cmd.transferlen=writechunk;
    the_cmd.buf=buf;
    the_cmd.timeout=30000l;
    retcode=scsi_cmd(&the_cmd,1);
    if(retcode)
      return retcode;
    buf+=writechunk;
    nblocks-=256;
    startblock+=256l;
  }
  if(nblocks) {
    par.Par.Address=startblock;
    par.Par.Length=(unsigned char)(nblocks&0xff);
    memcpy(the_cmd.cmdblk,par.asChar,sizeof(par));
    the_cmd.transferlen=(long)(nblocks*blocksize);
    the_cmd.buf=buf;
    the_cmd.timeout=30000l;
    retcode=scsi_cmd(&the_cmd,1);
  }
  return retcode;
}

int mode_select(short device, unsigned length, void *buf) {
  bzero(the_cmd.cmdblk,sizeof(the_cmd.cmdblk));
  the_cmd.cmdblk[0]=MODESELECT;
  the_cmd.cmdblk[4]=(unsigned char)length;
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  the_cmd.transferlen=length;
  the_cmd.buf=buf;
  the_cmd.timeout=200l;
  return scsi_cmd(&the_cmd,1);
}

int set_density(short device, long densitycode) {
  char densitybuf[12];
  long error;

  memset(densitybuf,0,12);
  densitybuf[3]=8;
  densitybuf[4]=0x81;
  *(long *)(densitybuf+8)=densitycode;
  error=mode_select(device,12,densitybuf);
  return error;
}

int mode_sense(short device, short PCF, short page,
	       unsigned length, void *buf) {
  bzero(the_cmd.cmdblk,sizeof(the_cmd.cmdblk));
  the_cmd.cmdblk[0]=MODESENSE;
  the_cmd.cmdblk[2]=((PCF&0x3)<<6)|(page&0x3f);
  the_cmd.cmdblk[4]=(unsigned char)length;
  the_cmd.device=device;
  the_cmd.cmdlen=6;
  the_cmd.transferlen=length;
  the_cmd.buf=buf;
  the_cmd.timeout=30000l;
  return scsi_cmd(&the_cmd,0);
}

int read_capacity(short device, void *buf) {
  bzero(the_cmd.cmdblk,sizeof(the_cmd.cmdblk));
  the_cmd.cmdblk[0]=0x25;		/* READCAPACITY */
  the_cmd.device=device;
  the_cmd.cmdlen=0x0a;
  the_cmd.transferlen=0x08l;
  the_cmd.buf=buf;
  the_cmd.timeout=30000l;
  return scsi_cmd(&the_cmd,0);
}

int scsi2atarierr(short device, long error) {
  if(!error)
    return 0;
  if((error&0x1e)==8) {
    return EDRVNR;
  }
  if((error&0x1e)==2) {
    SenseData sensedata;
    request_sense(device,&sensedata);
    if(sensedata.extendedsense.ErrorCode)
      return EDRVNR;	/* Signal error */
    switch(sensedata.extendedsense.SenseKey) {
    case 0:			/* No sense */
    case 1:			/* Recovered Error */
      return 0;
      break;
    case 2:
      return EDRVNR;
      break;
    case 3:
      return EREADF;
      break;
    case 4:
      return EREADF;
      break;
    case 5:
      return EBADRQ;
      break;
    case 6:
      return E_CHNG;
      break;
    case 7:
      return EACCDN;
      break;
    case 8:
      return EWRITF;
      break;
    case 9:
      return ERROR;
      break;
    case 10:
      return ERROR;
      break;
    case 11:
      return ERROR;		/* Should return E_TRY_AGAIN if that had */
      break;			/* existed... */
    case 12:
      return ERROR;
      break;
    case 13:
      return EWRITF;
      break;
    case 14:
      return ERROR;
      break;
    case 15:
      return ERROR;
      break;
    }
  }
  return ERROR;			/* Should not ever come here... */
}

char *sense_msg[15]={
  "NO SENSE",
  "Error Corrected",
  "LUN NOT READY",
  "MEDIUM ERROR",
  "HARDWARE ERROR",
  "ILLEGAL REQUEST",
  "UNIT ATTENTION (Media changed ?)",
  "DATA PROTECT",
  "BLANK CHECK",
  "Device Specific Sense Key",
  "Aborted COPY/COMPARE-Command",
  "EQUAL; SEARCH FOUND",
  "VOLUME OVERFLOW: Cannot write data in buffer",
  "MISCOMPARE",
  "RESERVED SENSE KEY"
  };

