#ifndef _SCSICMD_H
#define _SCSICMD_H

#include <sioctl.h>

#define CDTSTUNIT	0
#define CDREQSENSE	3
#define CDINQUIRE	0x12
#define CDREAD		8
#define CDSEEK		0xb
#define SCSI_LOCATE	0x2b
#define MODESELECT	0x15
#define MODESENSE	0x1a
#define STOPAUD		0xc6
#define STARTAUD	0x47
#define SETTIME		0x47
#define GETTOC		0x43
#define SUBQ		0x42
#define READDISC	0xC7		/* YYYY */
#define ZEROUNIT	0x01
#define STARTSTOP	0x1b

#define MAX_RETRY 10

#define CDROM_BLOCKSIZE 2048

#define  Load   1
#define  Unload 0

#define SPACE_BLOCKS		0
#define SPACE_FILEMARKS		1
#define SPACE_SEQFILEMARKS	2
#define SPACE_ENDOFDATA		3
#define SPACE_SETMARKS		4
#define SPACE_SEQ_SETMARKS	5

typedef struct {
  unsigned per_qual:3;
  unsigned per_dev_type:5;
  unsigned RMB:1;
  unsigned dev_modifier:7;
  unsigned iso_ver:2;
  unsigned ecma_ver:3;
  unsigned ansi_ver:3;
  unsigned AENC:1;
  unsigned TrmIOP:1;
  unsigned reserved1:2;
  unsigned resp_dataf:4;
  char add_len;
  char reserved2;
  char reserved3;
  unsigned RelAdr:1;
  unsigned WBus32:1;
  unsigned WBus16:1;
  unsigned Sync:1;
  unsigned Linked:1;
  unsigned reserved4:1;
  unsigned CmdQue:1;
  unsigned SftRe:1;
  char VendorID[8];
  char ProductID[16];
  char ProductRev[4];
} InquireData;

typedef struct {
  unsigned Valid:1;
  unsigned ErrorClass:3;
  unsigned ErrorCode:4;
  unsigned VendorSpecific:3;
  unsigned BlockMSB:5;
  unsigned short BlockLSB;
} NonExtendedSense;

typedef struct {
  unsigned Valid:1;
  unsigned ErrorClass:3;
  unsigned ErrorCode:4;
  unsigned char SegmentNumber;
  unsigned FM:1;
  unsigned EOM:1;
  unsigned ILI:1;
  unsigned Reserved:1;
  unsigned SenseKey:4;
  unsigned long InfoBytes;
  unsigned char ExtraBytes;
  unsigned long CommandSpec;
  unsigned char additionalSense;
  unsigned char additionalQualifier;
  unsigned char FieldReplaceableUnitCode;
  unsigned SKSV:1;
  unsigned long SenseSpecific:23;
} ExtendedSense;

typedef union {
  NonExtendedSense nonextendedsense;
  ExtendedSense extendedsense;
} SenseData;

typedef struct {
  unsigned char num_data;
  unsigned char medium_type;
  unsigned WP:1;
  unsigned devicespec:7;
  unsigned char block_descr_length;
  unsigned char DensityCode;
  unsigned long numBlocks:24;
  unsigned char reserved;
  unsigned long blockLength:24;
} SenseHeader;

typedef struct {
  unsigned reserved1:2;
  unsigned PageCode:6;
  unsigned char PageLength;
  unsigned long NumCylinder:24;
  unsigned char NumHeads;
  unsigned long WriteCompensation:24;
  unsigned long ReducedWriteStream:24;
  unsigned short Steprate;
  unsigned long LandingZone:24;
  unsigned char reserved2[3];
} DiskGeometry;

typedef struct {
  unsigned reserved:1;
  unsigned PageCode:7;
  unsigned short TracksPerZone;
  unsigned short AltSectorsPerZone;
  unsigned short AltTracksPerZone;
  unsigned short AltTracksPerLUN;
  unsigned short SectorsPerTrack;
  unsigned short BytesPerSector;
  unsigned short Interleave;
  unsigned short TrackSkewFactor;
  unsigned short CylinderSkewFactor;
  unsigned SSEC:1;
  unsigned HSEC:1;
  unsigned RMB:1;
  unsigned SURF:1;
  unsigned reserved2:4;
  unsigned char reserved3[3];
} FormatParameter;

typedef struct {
  unsigned long last_block;
  unsigned long block_length;
} DiskCapacity;

extern char *sense_msg[];

#include <scsiproto.h>
#endif /* _SCSICMD_H */
