#include <string.h>

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;
extern long status_timeout;
extern char *version;

union clong {
	char	aschar[4];
	long	aslong;
};

typedef struct cookie {
	union clong tag;
	long value;
} COOKIE;
#define CJAR	((COOKIE **) 0x5a0L)

#define TT		0x00020000
#define FALCON		0x00030000	/* Changed to reflect ROMs LT */

#include "filesys.h"
#include "atarierr.h"
#include "kmacros.h"
#include "steal.h"
#include "sioctl.h"
#include "scsidevproto.h"

/*
This is the plot:
   When a read/write request comes, we install interrupts, and 
   start dma. Then we add the process to the IO-queue with a
   sleep(), and install a  timeout for the requested delay.
   In the interrupt routine we install a function to be called in 0 ms.
   This means that the timeout will occur during the next VBL.
   In this (timeout-)function, we cancel the long timeout, 
   wake up the process, and remove the interrupt.
   To be sure that noone else tries to access the SCSI port
   while we are working, we grab the Rwabs vector.
   In our rwabs routine, we check if we are busy, if not
   we let the rwabs fall through to the usual handler.
   If indeed we are busy, we wait until the interrupt
   has been handled.

BUG:
   The "interrupt_done" variable is not a semaphore, and
   it *REALLY* should be !

   If a timeout occurs... Well, read the function, it should be
   obvious...
*/
   
short inittime,initdate;
struct kerinfo *kernel;
TIMEOUT *the_timeout=(TIMEOUT *)0l;
TIMEOUT *fake_timeout=(TIMEOUT *)0l;  
extern short sleeptime;

#define SCSI_SIMPLE 0
#define SCSI_READ   0
#define SCSI_WRITE  1

#define IO_Q 3

extern long do_scsi(short device, short ncmd,
			   long nbytes, short read,
			   char *cmdbuf, char *dmabuf);

DEVDRV scsi_dev={
  scsi_open,scsi_write,scsi_read,scsi_lseek,
  scsi_ioctl,scsi_datime,scsi_close,scsi_select,scsi_unselect
};

struct dev_descr devinfo = {
  &scsi_dev,0,0,(struct tty *)0l,
  {0l,0l,0l,0l}
  };

short status;
volatile long interrupt_done=1l;

long WAKEMAGIC =0l;

SELQUEUE *head=NULL, *tail=NULL;

SELQUEUE *mkqueue(long proc) {
  SELQUEUE *tmp=(SELQUEUE *)KMALLOC(sizeof(SELQUEUE));
  if(!tmp) {
    DEBUG(("mkqueue: Out of memory"));
    return NULL;
  }
  tmp->proc=proc;
  return tmp;
}

void insertqueue(SELQUEUE *entry) { /* Inserts entry at end */
  if(!head) {
    head=tail=entry;
    head->next=NULL;
    return;
  }
  tail->next=entry;
  entry->next=NULL;
  tail=entry;
}

void rmqueue(long proc) {
  SELQUEUE *tmp,*q;
  if(!head) {
    TRACE(("scsi rmqueue: Queue empty"));
    return;
  }
  if(head->proc==proc) {
    TRACE(("scsi rmqueue: removing head"));
    q=head;
    head=head->next;
    KFREE(q);
    if(!head) {
      TRACE(("scsi rmqueue: was only one entry, making tail=head=NULL"));
      tail=NULL;
    }
    return;
  }
  for(tmp=head;tmp;tmp=tmp->next) {
    if(tmp->next&&(tmp->next->proc==proc))
      break;
  }
  if(tmp&&tmp->next&&(tmp->next->proc==proc)) {
    TRACE(("scsi rmqueue: removing entry in the middle of list"));
    q=tmp->next;
    tmp->next=q->next;
    if(q==tail) {
      TRACE(("scsi rmqueue: well, not exactly middle - more like last"));
      tail=tmp;
    }
    KFREE(q);
    return;
  }
  ALERT(("Error: rmqueue didn\'t find entry for proc %lx",proc));
}

void wake_process(void) {
  long proc;
  TRACE(("System worked !!! waking up process...\r\n"));
  remove_interrupt();
  interrupt_done=1l;
  CANCELROOTTIMEOUT(the_timeout);
  WAKE(IO_Q, WAKEMAGIC);
  if(head) {
    proc=head->proc;
    rmqueue(proc);
    TRACE(("Waking up select()ing process %lx",proc));
    WAKESELECT(proc);
  }
}

void interrupt_routine(long stat) {
  status=(short)stat;
  fake_timeout=ADDROOTTIMEOUT(0l,wake_process,1); /* This is in an */
}				/* Interrupt, so set the "int" bit. */

extern long cleanup_interrupt(void);

void timeout_cleanup(void) {
  long proc;
  DEBUG(("Error: timeout\r\n"));
  status=(short)cleanup_interrupt();
  remove_interrupt();
  interrupt_done=1l;
  WAKE(IO_Q, WAKEMAGIC);
  if(head) {
    proc=head->proc;
    rmqueue(proc);
    TRACE(("Waking up select()ing process %lx",proc));
    WAKESELECT(proc);
  }
}
  
int install_interrupt(long timeout) {
  extern void set_mfp(void (*)(long)); /* From spscsi.cpp */
  the_timeout=ADDROOTTIMEOUT(timeout,timeout_cleanup,0);
  if(!the_timeout) {
    ALERT(("Unable to install timeout handler\r\n"));
    return 1;
  }
  set_mfp(interrupt_routine);
  return 0;
}

void remove_interrupt(void) {
  extern void reset_mfp(void);	/* From spscsi.cpp */
  reset_mfp();
}

long do_simple_command(SCSI_COMMAND *cmd) {
  long old_status_timeout=status_timeout;
  while(!interrupt_done) {
    DEBUG(("scsidev:Arggh ! Someone has used us without Select()ing :-( "));
    NAP(50);			/* Sleep 50ms before we check again */
  }
  TRACE(("scsi simple command"));
  interrupt_done=0l;
  status_timeout=cmd->timeout;
  cmd->cmdblk[1]|=((cmd->device&0x0700)>>3);
  cmd->status=(short)do_scsi(cmd->device&0x7,cmd->cmdlen,
				    0l,SCSI_SIMPLE, cmd->cmdblk,(char *)0l);
  status_timeout=old_status_timeout;
  interrupt_done=1l;
  return 0l;
}

long do_reading_command(SCSI_COMMAND *cmd) {
  long proc;
  long error;
  if(!interrupt_done)
    DEBUG(("scsidev:Arggh ! Someone has used us without Select()ing :-( "));
  while(!interrupt_done) {
    NAP(50);
  }
  interrupt_done=0l;
  WAKEMAGIC=(long)cmd;
  if(install_interrupt(cmd->timeout)) {
    cmd->status=ERROR;
    return 0l;
  }
  cmd->cmdblk[1]|=((cmd->device&0x0700)>>3);
  error=do_scsi(cmd->device&0x7,cmd->cmdlen,
		       cmd->transferlen,SCSI_READ, cmd->cmdblk,cmd->buf);
  if(error) {
    DEBUG(("Error while sending scsi command"));
    CANCELROOTTIMEOUT(the_timeout);
    interrupt_done=1l;
    cmd->status=(short)error;
    if(head) {
      proc=head->proc;
      rmqueue(proc);
      TRACE(("Waking select()ing process %lx",proc));
      WAKESELECT(proc);
    }
    return 0l;
  }
  SLEEP(IO_Q,WAKEMAGIC);
  if(!interrupt_done)
    if(PGETPID())
      DEBUG(("devscsi read: Interrupt is not done, why aren\'t we sleeping ?"));
  while(!interrupt_done) NAP(50);	/* This sometimes happens. Why ??? */
  cmd->status=status;		/* See comment at the end of writing_command */
  return 0l;
}

long do_writing_command(SCSI_COMMAND *cmd) {
  long error;
  long proc;
  if(!interrupt_done)
    DEBUG(("scsidev:Arggh ! Someone has used us without Select()ing :-( "));
  while(!interrupt_done) {
    NAP(50);
  }
  interrupt_done=0;
  WAKEMAGIC=(long)cmd;
  if(install_interrupt(cmd->timeout)) {
    cmd->status=ERROR;
    return 0l;
  }
  cmd->cmdblk[1]|=((cmd->device&0x0700)>>3);
  error=do_scsi(cmd->device&0x7,cmd->cmdlen,
		       cmd->transferlen,SCSI_WRITE, cmd->cmdblk,cmd->buf);
  if(error) {
    CANCELROOTTIMEOUT(the_timeout);
    cmd->status=(short)error;
    interrupt_done=1l;
    if(head) {
      proc=head->proc;
      rmqueue(proc);
      TRACE(("Waking up select()ing process %lx",proc));
      WAKESELECT(proc);
    }
    return 0l;
  }
  SLEEP(IO_Q,WAKEMAGIC);
  if(!interrupt_done)
    if(PGETPID())
      DEBUG(("devscsi write: Interrupt is not done, why aren\'t we sleeping ?"));
  while(!interrupt_done) NAP(50);	/* This sometimes happens. Why ??? */
  cmd->status=status;		/* I think I found out. If the process */
  return 0l;			/* trying to do io is process 0, that is */
}				/* MiNT itself, a SLEEP will immidieately */
				/* return, in order to avoid deadlocks. */
				/* I added the NAP, so that a context- */
				/* switch is possible, and the timeout/ */
				/* interrupt handler gets a chance to run.*/

long scsi_open(FILEPTR *f) {
  TRACE(("scsi_open(%lx)",f));
  f->pos=0;
  return 0;
}

long scsi_write(FILEPTR *f, const char *buf, long bytes) {
  return EINVFN;
}

long scsi_read(FILEPTR *f, char *buf, long bytes) {
  return EINVFN;
}

long scsi_lseek(FILEPTR *f, long where, int whence) {
  return EINVFN;
}

long scsi_ioctl(FILEPTR *f, int mode, void *buf) {
  TRACE(("scsi_ioctl(%lx,%d,%lx)",f,mode,buf));
  switch(mode) {
  case SSCMD:
    return do_simple_command((SCSI_COMMAND *)buf);
    break;
  case SRCMD:
    return do_reading_command((SCSI_COMMAND *)buf);
    break;
  case SWCMD:
    return do_writing_command((SCSI_COMMAND *)buf);
    break;
  case SSLEEP:
    sleeptime=(short)*(long *)buf;
    TRACE(("scsi_ioctl: Sleeptime %d",sleeptime));
    return 0l;
    break;
  case FIONREAD:
    *((long *)buf)=0l;
    return 0l;
  case FIONWRITE:
    *((long *)buf)=0l;
    return 0l;
  default:
    return EINVFN;
  }
}

long scsi_datime(FILEPTR *f, short *timeptr, int rwflag) {
  TRACE(("scsi_datime(%lx,%lx,%d)",f,timeptr,rwflag));
  if(rwflag)
    return EWRPRO;

  timeptr[0]=inittime;
  timeptr[1]=initdate;
  return 0l;
}

long scsi_close(FILEPTR *f, int pid) {
  return 0l;
}

long scsi_select(FILEPTR *f, long proc, int mode) {
  SELQUEUE *queEntry;
  
  TRACE(("scsi_select()"));
  switch(mode) {		/* Well, should setup a select-queue ! */
  case O_WRONLY:
    return 1;
  case O_RDONLY:
    if(((!head)&&(interrupt_done))||(!PGETPID())) /* Mustnt send */
      return 1;			/* MiNT itself to sleep ! */
    queEntry=mkqueue(proc);
    if(!queEntry){
      DEBUG(("Not enough memory for select-queue entry !!!"));
      return ENSMEM;
    }
    insertqueue(queEntry);
    DEBUG(("Sending process %lx to select()-sleep-heaven"));
    return 0;
    break;
  }
  DEBUG(("scsi_select: illegal mode"));
  return 1;
}

void scsi_unselect(FILEPTR *f, long proc, int mode) {
  TRACE(("scsi unselect: Removing %lx from select-queue",proc));
  rmqueue(proc);
  return;
}

extern void general_cleanup(void);

DEVDRV *scsi_init(struct kerinfo *k) {
  char *devname="u:\\dev\\scsi";
  COOKIE *jar;
  short machine=1;
  kernel=k;
  jar = *CJAR;
  if (jar) {
    while (jar->tag.aslong != 0) {
      if (!strncmp(jar->tag.aschar, "_MCH",4)) {
	if (jar->value == FALCON) {
	  machine = 1;
	} else if (jar->value == TT) {
	  machine = 0;
	} else {
	  machine = -1;
	}
	break;
      }
      jar++;
    }
  }
  CCONWS("\033pSCSI device driver\033q\r\nFalcon ");
  CCONWS(version);
  CCONWS("\r\n(c) Erling Henanger 1994\r\n");
  if(machine!=1) {
    CCONWS("This SCSI device driver only works on a Falcon\r\n");
    CCONWS("Unable to install scsi device\r\n");
  }
  if(ADDROOTTIMEOUT==0l) {
    CCONWS("The SCSI device driver needs a patched Mint 1.10\r\n");
    CCONWS("Unable to install scsi device\r\n");
    return (DEVDRV *)0l;
  }
  if(CANCELROOTTIMEOUT==0l) {
    CCONWS("The SCSI device driver needs a newer timeout patch for Mint 1.10\r\n");
    CCONWS("Unable to install scsi device\r\n");
    return (DEVDRV *)0l;
  }
  if(DCNTL(DEV_INSTALL,devname,&devinfo)<=0l) {
    CCONWS("Unable to install scsi device\r\n");
    return (DEVDRV *)0l;
  }
  inittime=TGETTIME();
  initdate=TGETDATE();
  (long)prev_rwabs   = *RWABS_Vector ;
  *RWABS_Vector   = (long)our_RWABS_vector ;

  CCONWS("SCSI device installed\r\n");
  return (DEVDRV *)1l;
}

long  our_RWABS_vector(int rwflag, void *buf, int cnt,
		       int recno, int dev, long lrecno ) {
  long result ;			/* We might do some checking. */
				/* If the device in question is */
				/* the IDE disk, we could let the */
				/* RWABS call fall through even if */
				/* we are busy. */
  if(!interrupt_done)
    DEBUG(("scsidev:Bad boy program tried to do rwabs while we were busy !!"));
  while(!interrupt_done) NAP(100); /* Dirty, I know... */
  result = Bios_RWABS(rwflag, buf, cnt, recno, dev, lrecno) ;
  return result ;
}
