#include "spconst.h"
#include "spmacro.h"

#define timer 0x4ba:w
#define flock 0x43e:w
#ifdef DEBUG
#define CCONWS(a)  	moveml d0-d2/a0-a2,sp@-; \
		    	pea a; \
		 	movew \#9,sp@-; \
			trap \#1; \
			addql \#6,sp; \
			moveml sp@+,d0-d2/a0-a2 
#define PRINTWORD(a)	movew a,d0; bsr printword
#else
#define CCONWS(a)
#define PRINTWORD(a)
#endif
#define SUPER(a)	movel a,sp@-;movew \#0x20,sp@-;trap \#1;addql \#6,sp

short_timeout		= 101		| > 0.5 sec
long_timeout 		= 201		| >   1 sec
extra_long_timeout 	= 1001		| >   5 sec


#ifdef DEBUG
okselstring:	.ascii	"Selected device ok\015\012\000"
okcmdstring:	.ascii	"Sent command ok\015\012\000"
selstring:	.ascii	"About to select device \000"
ackstring:	.ascii	"Doing ack\015\012\000"
writebytestr:	.ascii	"Doing write_byte\015\012\000"
waitforints:	.ascii	"Doing wait for interrupt\r\n\000"
numbers:	.ascii	"0123456789ABCDEF\000"
getstatstr:	.ascii	"Doing get_stat\r\n\000"
dmareadstr:	.ascii	"Starting read  DMA\r\n\000"
dmawritestr:	.ascii	"Starting write DMA\r\n\000"
crlf:		.ascii	"\015\012\000"
#endif

_status_timeout:	.long	long_timeout
old_timer:		.long	0	| Value for timer from last time
cpu030:			.byte	1	| Yes, we are a 68030
flush:			.byte	0	| Flag: Do we need to flush the fifo ?
_sleeptime:		.word	0	| How long should we sleep between
					| commands ? Must be set externally
old_interrupt:		.long	0	| Pointer to old DMA interrupt routine
interrupt_func:		.long	0	| Function to call in interrupt
interrupt_installed:	.byte	0	| Flag that tells whether
					| interrupt is installed or not.
oldstack:		.long	0
transferlength:		.long	0
_poll_timeout:		.long   200
status_reg:		.word	0

DMA_INT = 0x11c

		.globl  _resetspscsi	
		.globl	_do_scsi
		.globl	_do_scsi32
		.globl	_sleeptime
		.globl	_status_timeout
		.globl  _poll_timeout
#ifdef CHECK_DISKINTERRUPT
		.globl	_in_diskinterrupt
#endif
		.globl	testFlock
		.globl	do_delay
		.globl	general_cleanup
		.globl	_spl7
		.globl	_splnormal
#ifdef CHECK_SUPER
		.globl	_in_super
#endif
		.even
		.text

_spl7:
#ifdef CHECK_SUPER
	tstl	_in_super
	bnes	setit
	moveml	d1-d2/a0-a2,sp@-
	clrl	sp@-
	movew	#0x20,sp@-
	trap	#1
	addql	#6,sp
	movel	d0,oldstack
	moveml	sp@+,d1-d2/a0-a2
setit:
#endif
	movew	sr,d0
	movew	d0,status_reg
	oriw	#0x0700,sr
	rts

_splnormal:
	movew	status_reg,d0
	movew	d0,sr
#ifdef CHECK_SUPER
	tstl	_in_super
	beqs	leave_super
	rts
leave_super:
	moveml	d1-d2/a0-a2,sp@-
	movel	oldstack,sp@-
	movew	#0x20,sp@-
	trap	#1
	addql	#6,sp
	moveml	sp@+,d1-d2/a0-a2
	moveql	#0,d0
#endif
	rts
	
	.globl	_set_mfp		| Enables interrupt, and installs
					| func to be called during DMAint.
_set_mfp:				| set_mfp(void (*func)(long status))
	st	interrupt_installed	| Yes, interrupt routine is installed.
	movel	sp@(4),interrupt_func	| Get interrupt function.
	movel	DMA_INT, old_interrupt  | REAL CODE WOULD TAKE CARE OF
					| ALL THESE REGS, AND RESTORE THEM
					| LATER. MIGHT CLASH WITH PREVIOUSLY
					| INSTALLED DMA INTERRUPTS !!!
	bclr	#7, IERB		| Disable interrupt while initing
					| Should really goto ipl7()....
	bclr	#7, IMRB		| And mask out interrupts as well 
	movel	#DMAinterrupt, DMA_INT	| Install our handler
	bclr	#5, AER			| Interrupt on high->low transition
	bclr	#7, IPRB		| Clear any pending interrupts
	bset	#7, IMRB		| Interrupt mask register   B
	bset	#7, IERB		| Interrupt enable register B
	rts

	.globl	_reset_mfp

_reset_mfp:
	tstb	interrupt_installed
	beqs	notinstalled
	sf	interrupt_installed
					| SEE NOTE ABOUT RESTORING 
					| REG VALUES FROM INSTALL_INT !
	bclr	#7, IERB		| Disable interrupt
	bclr	#7, IMRB		| Mask interrupt
	bclr	#7, IPRB		| Clear any pending interrupts
	movel	old_interrupt, DMA_INT	| Restore old interrupt handler
notinstalled:
	rts

	.globl _cleanup_interrupt
_cleanup_interrupt:
	RSCSI	(#s_inircv,d0)		| clear potential interrupt
	WSCSI	(#0,#s_mode)		| disable DMA mode
	WSCSI	(#0,#s_icr)		| make sure data bus is not asserted

	tstb	flush			| need to flush FIFO?
	beqs	rint3			| if not, proceed	
	movew	#0x0020,DMA_MODE	| else flush (Whoosh!)
|-
rint3:	
	tstb	cpu030			| is there a cpu cache to flush?
	beqs	int_getstat		| if not, go get status
	movew	sr,sp@-			| go to IPL 7
	oriw	#0x700,sr		| no interrupts right now kudasai
	movec 	cacr,d0			| d0 = (cache control register)
	oriw	#0x808,d0		| dump both the D and I cache
	movec	d0, cacr		| Update cache control register
	movew	sp@+,sr			| restore interrupt state
int_getstat:
	bsr	get_stat
	RSCSI	(#s_inircv,d1)		| clear potential interrupt
	movew	#0x80,0x8606:w
	clrw	flock
#ifdef CHECK_DISKINTERRUPT
	clrl	_in_diskinterrupt
#endif
	rts

DMAinterrupt:
	moveml	d0-d2/a0-a2,sp@-
	bsrs	_cleanup_interrupt	| Get status, etc.
	bclr	#7, ISRB		| inform MFP of interrupt end
	movel	d0,sp@-			| Call the "C" interrupt handler
	movel	interrupt_func,a0
	jsr	a0@
	addql	#4,sp
	bsr	general_cleanup		| And do some general cleanup
	moveml	sp@+,d0-d2/a0-a2	| Keep some instruction between
					| moveml and rte because of bug in
					| 68000
	rte

testFlock:
	moveml	d1/a0,sp@-
	moveql	#0,d0			| Returner 0 som default
	movel	timer,d1
	addl	extra_long_timeout,d1
flockLoop:
	tas	flock
	beqs	flockExit
	cmpl	timer,d1
	bccs	flockLoop
	moveql	#-8,d0			| Error, timeout
flockExit:
	moveml	sp@+,d1/a0
	rts

do_delay:
	tstw	_sleeptime
	beqs	justexit
	movel	d0,sp@-
	movel	old_timer,d0
delayagain:
	cmpl	timer,d0		| Wait according to delaysettings
	bpls	delayagain
	movel	sp@+,d0
justexit:
	rts

		.globl	_general_cleanup
_general_cleanup:
general_cleanup:
	movel	d0,sp@-
	moveql	#0,d0
	movew	_sleeptime,d0
	addl	timer,d0		| Wait at least "_sleeptime" * 5 ms
	movel	d0,old_timer		| between two successive commands
	movel	sp@+,d0
	rts

_do_scsi32:	|(long device, long ncmd, long nbytes, long read,
			| char *cmdbuf, char *dmabuf)

#ifdef CHECK_DISKINTERRUPT
	tas	_in_diskinterrupt
	bnes	_do_scsi32
#endif
	movel	sp, d3
	moveml	d1-d7/a0-a6,sp@-

#ifdef FROM_USER_MODE
	SUPER	(#0)
	movel	d0,oldstack
#endif
	movel	d3,a0
	movel	a0@(4),d0	| device 		short
	movel	a0@(8),d2	| ncmd			short
	movel	a0@(12),d1	| nbytes		long
	movel	a0@(16),d3	| read/write		short
	movel	a0@(20),a1	| Command buffer	long
	movel	a0@(24),a2	| Dma buffer		long
	bras	got_regs

_do_scsi:		|(short device, short ncmd, long nbytes,short read,
			| char *cmdbuf, char *dmabuf)
	movel	sp, d3
	moveml	d1-d7/a0-a6,sp@-

#ifdef FROM_USER_MODE
	SUPER	(#0)
	movel	d0,oldstack
#endif
	movel	d3,a0
	movew	a0@(4),d0	| device 		short
	movew	a0@(6),d2	| ncmd			short
	movel	a0@(8),d1	| nbytes		long
	movew	a0@(12),d3	| read/write		short
	movel	a0@(14),a1	| Command buffer	long
	movel	a0@(18),a2	| Dma buffer		long
got_regs:
	lea	DMA_ACCESS,a0	| Dma chip
	tstl	d1		| Are we transferring anything ?
	bnes	do_real_command | Yes
	bsr	simple_command
	bras	restore_regs
do_real_command:
	bsr	scsi_command
restore_regs:

#ifdef FROM_USER_MODE
	movel	d0,sp@-
	SUPER	(oldstack)
	movel	sp@+,d0
#endif
	moveml	sp@+,d1-d7/a0-a6
	rts
	
#ifdef AHDI_COMPAT
		.globl	_cmdblk
		.globl	rcvspscsi
		.globl	wrtspscsi
		.globl	smplspscsi
	| AHDI compatibility stuff...
	| 	d0 = device
	|	d1 = transfer length
	| 	d2 = commandlength
	|	d3 = read or write
	|	a0 = dmachip
	| 	a1 = commandblock adresse
	|	a2 = dma buffer

rcvspscsi:
	movel	a0,a2		| This is where *we* get the dmabuffer from
	lea	DMA_ACCESS,a0
	lea	_cmdblk,a1
	moveql	#0,d3		| Setup for reading
	bras	scsi_command

wrtspscsi:
	movel	a0,a2
	lea	DMA_ACCESS,a0
	lea	_cmdblk,a1
	moveql	#1,d3		| Setup for writing
	bras	scsi_command


				| d0 = device
				| d1 = nbytes in dmabuffer
				| d2 = command length
				| a1 = cmdblk address
smplspscsi:
	lea	_cmdblk,a1
#endif /* AHDI_COMPAT */

simple_command:
	bsr	do_delay
	bsr	send_block
	bmis	exit_simple
	bsr	get_stat
exit_simple:
	bsr	_hdone
	bsr	general_cleanup		| Save timer and stuff
	rts		

	|	d1 = transfer length
	|	a2 = pointer to write buffer

scsi_command:
	movel	d1,transferlength
	bsr	do_delay
	andl	#1,d3			| Only bit one counts - read or write.
	moveml	d1-d2,sp@-		| Save device, command length
	bsr	send_block
	moveml	sp@+,d1-d2		| Restore device, command length

	tstl	d0
	bmis	end_scsi_command
	bsr	start_dma
	asrl	#8,d1
	asrl	#1,d1
|	addl	#extra_long_timeout,d1
	addl	_poll_timeout,d1
	addl	timer,d1
	tstl	d0
	beqs	do_polling

programmed_write:			| Continue until no longer phasematch
	bsr	get_request		| OK, following this comes
					| all the funky stuff.
	bmis	end_scsi_command	| Sorry about the lack of comments,
	bsr	check_phasematch	| but if you know anything about
	beqs	done_writing		| SCSI programming, the code is pretty
	moveb	a2@+,d0			| obvious. Go and get the SCSI2
	bsr	write_byte		| specification, available by ftp
	bsr	do_acknowledge		| from ftp.cdrom.com (I think it was)
	bras	programmed_write	| Check "Profibuch" or ataris
do_polling:				| special docs about the NCR chip.
	tstb	interrupt_installed	| (If they exist, I only got some
	beqs	stupid_polling		| source code from AHDI...)
	rts				| erlingh@pvv.unit.no
stupid_polling:
	bsr	wait_for_ints
	tstl	d0
	bnes	end_scsi_command
done_writing:
	bsr	get_stat
	tstl	d0
	bnes	end_scsi_command
	moveql	#0,d0
end_scsi_command:
	bsr	_hdone
	bsr	general_cleanup
	rts
		
	| d0 er device
	| d2 er commandlength
	| a1 er commandblock adresse

send_block:	
	movew	d0,sp@-		| push device
	bsr 	select_device
	addql	#2,sp
	tstl	d0		| Select ok ?
	bmis	end_block	| nope
	bsr	assert_cd_data
	movel	#long_timeout,d1
	addl	timer,d1
	subqw	#1,d2
nextbyte:
	moveql	#0,d0
	moveb	a1@+,d0
	bsr	write_byte
	tstl	d0
	bmis	error_block
	dbra	d2,nextbyte
	moveql	#0,d0
end_block:
	rts
error_block:
	movel	#-7,d0
	rts
			| sp@(4).w == device
select_device:
|	st	flock
	bsr	testFlock
	movel	d0,d1
	bnes	only_exit_select
	movel	#long_timeout,d1
	addl	timer,d1
clear_again:
	bsr	read_idstat
	btst	#6,d0
	beqs	got_busy_clear
	cmpl	timer,d1
	bhis	clear_again
	movel	#-1,d1
	bras	exit_select
got_busy_clear:
	bsr	assert_busysel
	movew	sp@(4),d1	| Get device no
	bsr	do_select
	movel	#long_timeout,d1
	addl	timer,d1
check_busy:
	bsr	read_idstat
	btst	#6,d0
	bnes	okbusy
	cmpl	timer,d1
	bhis	check_busy
	movel	#-2,d1
	bras	exit_select
okbusy:
	moveql	#0,d1
exit_select:
	bsrs	clear_icr
only_exit_select:
	movel	d1,d0
	rts

clear_icr:
	WSCSI	(#0,#s_icr)
	rts

assert_busysel:
	WSCSI	(#0,#s_tcr)
	WSCSI	(#0,#s_idstat)
	WSCSI	(#0xc,#s_icr)
	rts

do_select:				| d1 = device
	CCONWS	(selstring)
	PRINTWORD(d1)
	CCONWS	(crlf)

	moveql	#0,d0
	andl	#7,d1
	bset	d1,d0

	WSCSI	(d0,#s_data)
	WSCSI	(#5,#s_icr)
	RSCSI	(#s_mode,d0)
	andw	#0xfe,d0
	movew	d0,DMA_ACCESS
	RSCSI	(#s_icr,d0)
	andw	#0xf7,d0
	movew	d0,DMA_ACCESS
	rts

read_idstat:
	RSCSI	(#s_idstat,d0)
	andw	#0xff,d0
	rts

read_icr:
	RSCSI	(#s_icr,d0)
	andw	#0xff,d0
	rts

assert_cd_data:
	WSCSI	(#2,#s_tcr)
	WSCSI	(#1,#s_icr)
	rts
				| Data to be written is in d0
write_byte:
	movew	d0,sp@-
	bsr	get_request
	bmis	error_write_byte
	movew	sp@+,d0
	bsrs	do_writebyte
	bsr	do_acknowledge
	rts
error_write_byte:
	tstw	sp@+
	movel	#-6,d0
	rts

do_writebyte:
	WSCSI	(d0,#s_data)
	rts

reset_and_exit:
	bsrs	reset_scsi
	movel	#-3,d0
	rts

wait_for_ints:
	CCONWS	(waitforints)
checkmfp:
	btst	#5,GPIP
	beqs	got_int
	cmpl	timer,d1
	bhis	checkmfp
	bras	reset_and_exit
got_int:
	RSCSI	(#s_inircv,d1)
	WSCSI	(#0,#s_mode)
	WSCSI	(#0,#s_icr)
	tstl	d3			| Are we reading or writing ?
	bnes	exit_wait_for_ints	| Writing, so we dont need to flush
	tstb	flush			| Do we need to flush the FIFO ?
	beqs	check_cpu
	movew	#0x20, DMA_MODE		| Flush it
check_cpu:
	tstb	cpu030			| Is this a 68030 with processor
					| cache ?
	beqs	exit_wait_for_ints	| Nope
	bsr	flush_cpu_cache		| Yes, flush cpu !
exit_wait_for_ints:
	moveql	#0,d0
	rts

flush_cpu_cache:
	movew	sr,sp@-
	oriw	#0x700,sr
	movec	cacr,d0
	oril	#0x808,d0
	movec	d0,cacr
	movew	sp@+,sr
	rts

_resetspscsi:
reset_scsi:
	btst	#3,0x860f:w	| Dma chip safe to access ?
	bnes	reset_scsi
	WSCSI	(#0x80,#s_icr)
	movel	#short_timeout,d1
	addl	timer,d1
waitsomemore:
	cmpl	timer,d1
	bhis	waitsomemore
	WSCSI	(#0,#s_icr)
	movel	#extra_long_timeout,d1
	addl	timer,d1
waitevenmore:
	cmpl	timer,d1
	bhis	waitevenmore
	rts

start_dma:
	tstl	d3
	beqs	reading
	CCONWS	(dmawritestr)
	bras	startit
reading:
	CCONWS	(dmareadstr)	| d3 == 0, read:d3 == 1 write
startit:
	sf	flush		| Assume we dont need to flush
	movel	d1,d0
	andl	#0xf,d0		| Multiple of 16 bytes to read/write ?
	beqs	dont_flush	| Yes
	st	flush		| Nope, we need to flush the fifo
dont_flush:
	tstl	d3
	bnes	skip_icr
	WSCSI	(#0, #s_icr)
skip_icr:
	moveql	#1,d0
	eorb	d3,d0
	WSCSI	(d0,#s_tcr)
	RSCSI	(#s_inircv,d0)
	moveql	#1,d0
	tstl	d3
	beqs	do_dma	
	cmpl	#0x200,d1	| less than 512 bytes ?
	bcss	just_exit_dma
do_dma:
	WSCSI	(#2,#s_mode)
	movel	a2,sp@-		| set dma address
	moveb	sp@(3),0x860d:w	| DMA lo
	moveb	sp@(2),0x860b:w	| DMA mid
	moveb	sp@(1),0x8609:w	| DMA hi
	addql	#4,sp
	movel	d3,sp@-		| Save read/write flag
	lsll	#8,d3		| Write flag into bit 8
	bsr	toggle_dma_chip
	movel	sp@+,d3		| Restore read/write flag
	movel	d1,sp@-
	addil	#0x1ff,d1	| Save nbytes
	asrl	#8,d1		| nbytes=(nbytes+511)/512 == n 512 byte blocks
	asrl	#1,d1
	movew	d1,a0@		| Write sector count register
	movel	sp@+,d1		| Restore nbytes
safe_dma:
	btst	#3,0x860f:w	| Dma chip safe to access ?
	bnes	safe_dma
	tstl	d3		| reading or writing ?
	beqs	start_readdma
start_writedma:
	movew	#0x18d,DMA_MODE
	movew	#0,DMA_ACCESS
	movew	#0x100,a0@(2)
	bras	exit_dma
start_readdma:
	WSCSI	(#0, #s_inircv)
	movew	#0,a0@(2)
exit_dma:
	moveql	#0,d0
just_exit_dma:
	rts

toggle_dma_chip:
	movew	#0x190,d0
	eorw	d3,d0			| XOR the write bit (bit 8)
	movew	d0,a0@(2)
	bsr	waitsome
	movew	#0x90,d0
	eorw	d3,d0			| And XOR again...
	movew	d0,a0@(2)
	bsr	waitsome
	rts

waitsome:
	tstb	0xfa01:w
	tstb	0xfa01:w
	tstb	0xfa01:w
	tstb	0xfa01:w
	rts

get_stat:
	CCONWS	(getstatstr)
	movel	_status_timeout,d1
	addl	timer,d1
	bsr	set_status_in_phase
	bsr	get_request
	bmis	error_status
	bsr	get_data
	movel	d0,sp@-			| Push status byte
	bsr	do_acknowledge
	tstl	d0
	bnes	error_status
	movel	#long_timeout,d1
	addl	timer,d1
	bsr	get_request
	bmis	error_status
	bsr	get_data
	bsr	do_acknowledge
	tstl	d0
	bmis	error_status
	movel	sp@+,d0			| Recall status byte
	rts
error_status:
	addql	#4,sp			| Return error
	rts
	
do_acknowledge:
	RSCSI	(#s_icr,d0)
	orib	#0x11,d0
	movew	d0,DMA_ACCESS
	movew	#s_idstat,DMA_MODE
req_loop:
	movew	DMA_ACCESS,d0
	btst	#5,d0
	beqs	req_clear
	cmpl	timer,d1
	bhis	req_loop
	movel	#-4,sp@-
	bras	ack_end
req_clear:
	movel	#0,sp@-
ack_end:
	RSCSI	(#s_icr,d0)
	andib	#0xef,d0
	movew	d0,DMA_ACCESS
	movel	sp@+,d0
	rts

get_data:
	RSCSI	(#s_data,d0)
	andiw	#0xff,d0
	rts

get_request:
	movew	#s_idstat,DMA_MODE
req_again:
	movew	DMA_ACCESS,d0
	btst	#5,d0
	bnes	exit_request
	cmpl	timer,d1
	bhis	req_again
error_request:
	bsr	reset_scsi
	movel	#-5,d0
	rts
exit_request:
	moveql	#0,d0
	rts

set_status_in_phase:
	WSCSI	(#3,#s_tcr)
	RSCSI	(#s_inircv,d0)
	rts

check_phasematch:
	RSCSI	(#s_dmastat,d0)
	btst	#3,d0
	rts

_hdone:
	RSCSI	(#s_inircv,d1)
	movew	#0x80,DMA_MODE
#ifdef CHECK_DISKINTERRUPT
	clrl	_in_diskinterrupt
#endif
	clrw	flock
	rts

#ifdef DEBUG
bconout:
	moveml	d1-d2/a0-a2,sp@-
	movew	d0,sp@-
	movew	#2,sp@-
	movew	#3,sp@-			| Bios(3) == Bconout(2,d0)
	trap	#13
	addql	#6,sp
	moveml	sp@+,d1-d2/a0-a2
	rts

printbyte:
	moveml	d1-d2/a0,sp@-
	lea	numbers,a0
	andl	#0xff,d0
	movew	d0,d1
	movew	d1,d2
	lsrw	#4,d1
	moveb	a0@(d1:w),d0
	bsr	bconout
	andw	#0xf,d2
	moveb	a0@(d2:w),d0
	bsr	bconout
	moveml	sp@+,d1-d2/a0
	rts

printword:|(d0:w)
	movel	d1,sp@-
	andl	#0xffff,d0
	movew	d0,d1
	lsrw	#8,d0
	bsr	printbyte
	movew	d1,d0
	andw	#0xff,d0
	bsr	printbyte
	movel	sp@+,d1
	rts

printlong:|(d0:l)
	movel	d1,sp@-
	movel	d0,d1
	swap	d0
	bsr	printword
	movel	d1,d0
	bsr	printword
	movel	sp@+,d1
	rts
#endif
