#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <macros.h>
#include <osbind.h>
#include <mintbind.h>
#include "sioctl.h"

char *progname;
char devicename[80]="u:\\dev\\tape";

void usage(void) {
  printf("%s: [-d devname] -r -mt -um -sb nblocks -sf nblocks -e [0/1]\n",
	 progname);
  printf("-r  == rewind  tape\n");
  printf("-mt == mount   tape\n");
  printf("-um == unmount tape\n");
  printf("-sb == skip    blocks (can take negative arg)\n");
  printf("-sf == skip    filemark\n");
  printf("-e  == erase   longflag\n");
  exit(1);
}

void do_erase(int longflag) {
#if 0
  error=scsi_erase(device,0 /* Immed */, longflag,
		   (longflag)?300000l:10000l);
  if(error) {
    printf("Error while erasing\n");
    explain_error(device,error);
    exit(1);
  }
  printf("Erase: OK\n");
#else
  printf("Erase: Not implemented\n");
#endif
}

void do_mount(void) {
  long error;
  int fd=Fopen(devicename,0);
  if(fd<0) {
    fprintf(stderr,"Could not open device %s\n",devicename);
    exit(1);
  }
  error=Fcntl(fd,0l,TMOUNT);
  Fclose(fd);
  if(error) {
    fprintf(stderr,"Error %ld mounting %s\n",error,devicename);
    exit(1);
  }
  printf("Mount %s OK\n",devicename);
}

void do_umount(void) {
  long error;
  int fd=Fopen(devicename,0);
  if(fd<0) {
    fprintf(stderr,"Could not open device %s\n",devicename);
    exit(1);
  }
  error=Fcntl(fd,0l,TUMOUNT);
  Fclose(fd);
  if(error) {
    fprintf(stderr,"Error %ld unmounting %s\n",error,devicename);
    exit(1);
  }
  printf("Unmount %s OK\n",devicename);
}

void do_rewind(void) {
  long error;
  int fd=Fopen(devicename,0);
  if(fd<0) {
    fprintf(stderr,"Could not open device %s\n",devicename);
    exit(1);
  }
  error=Fcntl(fd,0l,TREW);
  Fclose(fd);
  if(error) {
    fprintf(stderr,"Error %ld rewinding %s\n",error,devicename);
    exit(1);
  }
  printf("Rewind %s OK\n",devicename);
}

void do_skip(long nblocks) {
  long error;
  int fd=Fopen(devicename,0);
  if(fd<0) {
    fprintf(stderr,"Could not open device %s\n",devicename);
    exit(1);
  }
  error=Fcntl(fd,&nblocks,TSFILEM);
  Fclose(fd);
  if(error) {
    fprintf(stderr,"Error %ld skipping %ld filemarks on %s\n",
	    error,nblocks,devicename);
    exit(1);
  }
  printf("Skip %ld filemarks on %s OK\n",nblocks,devicename);
}

void skip_blocks(long nblocks) {
#if 0
  printf("Skipping %ld blocks\n",nblocks);
  fflush(stdout);
  error=scsi_space(device,nblocks,SPACE_BLOCKS, abs(nblocks)*400l+30000l);
  if(error) {
    printf("Skip blocks failed\n");
    explain_error(device,error);
    exit(1);
  }
  printf("Skip blocks OK\n");
#else
  printf("Skip blocks unimplemented\n");
#endif
}

int main(int argc, char *argv[]) {
  int rewind=0,mount=0,unmount=0,skip=0,skipblocks=0,erase=0;
  int i;
  long nblocks=0l;
  int longflag=0;

  progname=argv[0];

  if(argc<2)
    usage();

  for(i=1;i<argc;i++) {
    if(!strcmp(argv[i],"-d")) {
      strcpy(devicename,argv[i+1]);
      i++;
      continue;
    }
    if(!strcmp(argv[i],"-sb")) {
      nblocks=atol(argv[i+1]);
      skipblocks=1;
      i++;
      continue;
    }
    if(!strcmp(argv[i],"-e")) {
      longflag=atoi(argv[i+1]);
      i++;
      erase=1;
      continue;
    }
    if(!strcmp(argv[i],"-r")) {
      rewind=1;
      continue;
    }
    if(!strcmp(argv[i],"-mt")) {
      mount=1;
      continue;
    }
    if(!strcmp(argv[i],"-um")) {
      unmount=1;
      continue;
    }
    if(!strcmp(argv[i],"-sf")) {
      skip=1;
      nblocks=atol(argv[i+1]);
      i++;
      continue;
    }
    printf("%s: Illegal option %s\n",progname,argv[i]);
    usage();
  }
  if(rewind+mount+unmount+skip+skipblocks+erase!=1) {
    printf("%s: Exactly one of the options:\n\t-r -e -mt -um -sb -sf\nmust be specified\n",
	   progname);
    return 1;
  }
  if(erase)
    do_erase(longflag);
  if(skipblocks)
    skip_blocks(nblocks);
  if(rewind)
    do_rewind();
  if(mount)
    do_mount();
  if(unmount)
    do_umount();
  if(skip)
    do_skip(nblocks);
  return 0;
}
