#include <macros.h>
#include <string.h>
#include <osbind.h>
#include <mintbind.h>
#include <signal.h>
#include <basepage.h>

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;

#include "scsicmd.h"
#include "filesys.h"
#include "atarierr.h"
#include "kmacros.h"
#include "tapeproto.h"

#define TIMEOUT 120000l		/* 2 min. timeout */

struct options {
  char  mountpoint[32];
  short device;
  short lun;
  struct {
    unsigned test_unit_ready:1;
    unsigned load_unload:1;
    unsigned lock:1;
    unsigned mode_sense:1;
    unsigned mode_select:1;
    unsigned immed:1;
    unsigned reserved:2;
  } Flags;
  short block_size;
  short timeouts[5];
} Options = {
  "u:\\dev\\tape",
  6,0,{
    0,
    1,
    1,
    1,
    0,
    1,
    0
    },
      512, { 1000,1100,1200,1300,1400 }
};

char dmabuf[8192];		/* Lets just hope that the block length */
short inittime,initdate;	/* isnt more than this... */
struct kerinfo *kernel;
extern char *version;

long blocksize=1024l;

DEVDRV tape_dev={
  tape_open,tape_write,tape_read,tape_lseek,
  tape_ioctl,tape_datime,tape_close,tape_select,tape_unselect
};

struct dev_descr devinfo = {
  &tape_dev,0,0,(struct tty *)0l,
  {0l,0l,0l,0l}
  };

void load_config(void) {
  int fd;
  fd=Fopen("u:\\multitos\\tapedev.cfg",0);
  if(fd<0)
    fd=Fopen("u:\\mint\\tapedev.cfg",0);
  if(fd<0)
    fd=Fopen("c:\\multitos\\tapedev.cfg",0);
  if(fd<0)
    fd=Fopen("c:\\mint\\tapedev.cfg",0);
  if(fd<0)
    fd=Fopen("c:\\tapedev.cfg",0);
  if(fd<0)
    return;
  Fread(fd,sizeof(Options),&Options);
  Fclose(fd);
}

int read_block(int device,long nblocks, char *buf) {
  int retcode;
#ifndef NO_DEBUG
  extern char *sense_msg[];
#endif

  if(Options.Flags.test_unit_ready)
    if((retcode=test_unit_ready(device)))
      goto jumphere;

again:
  TRACE(("read_block(%d, %ld, %lx), Options.block_size %ld",device,nblocks,buf,Options.block_size));
  if((nblocks==1)&&(buf==NULL)) {
    retcode=scsi_seqread(device,1l,Options.block_size,dmabuf,
			 (Options.timeouts[4]+Options.timeouts[3])*200l);
    DEBUG(("Returning from scsi_seqread"));
  } else {
    DEBUG(("read_block:  nblocks %ld, buf %06lx",
	   nblocks,buf));
    retcode=scsi_seqread(device,nblocks,Options.block_size,buf,
			 (Options.timeouts[4]*nblocks+Options.timeouts[3])*200l);
    DEBUG(("Returning from scsi_seqread"));
  }
 jumphere:
  if(retcode) {
    DEBUG(("read_block: scsi_read returned %d",retcode));
    if((retcode&0x1e)==8) {
      DEBUG(("read_block: Device BUSY"));
      return EDRVNR;		/* Return Drive Not Ready */
    } else if((retcode&0x1e)==2) { /* OK, we will request sense... */
      SenseData sensedata;
      request_sense(device, &sensedata);
      TRACE(("tapedev writeblock, requesting sense"));
      request_sense(device, &sensedata);
      DEBUG(("Valid: %d, FM: %d, EOM: %d, ILI: %d, ErrClass %d, ErrCode %d",
	     sensedata.extendedsense.Valid,
	     sensedata.extendedsense.FM,
	     sensedata.extendedsense.EOM,
	     sensedata.extendedsense.ILI,
	     sensedata.extendedsense.ErrorClass,
	     sensedata.extendedsense.ErrorCode));
      DEBUG(("SegmentNumber %d, InfoBytes %lx, ExtraBytes %d",
	     sensedata.extendedsense.SegmentNumber,
	     sensedata.extendedsense.InfoBytes,
	     sensedata.extendedsense.ExtraBytes));
      DEBUG(("CommandSpec: %lx, additionalSense %d, additionalQualifier %d",
	     sensedata.extendedsense.CommandSpec,
	     sensedata.extendedsense.additionalSense,
	     sensedata.extendedsense.additionalQualifier));
      DEBUG(("FieldReplacableUnitCode %d, SKSV %d, SenseSpecific %lx",
	     sensedata.extendedsense.FieldReplaceableUnitCode,
	     sensedata.extendedsense.SKSV,
	     sensedata.extendedsense.SenseSpecific));
      if(sensedata.nonextendedsense.ErrorClass<7) {
	ALERT(("read_block: Nonextended sense: class %u, code %u",
	      sensedata.nonextendedsense.ErrorClass,
	      sensedata.nonextendedsense.ErrorCode));
	if(sensedata.nonextendedsense.Valid)
	  ALERT(("read_block: in block %lu",
		((long)sensedata.nonextendedsense.BlockMSB<<16)|
		sensedata.nonextendedsense.BlockLSB));
	return EDRVNR;
      } else if(!sensedata.extendedsense.ErrorCode) {
	DEBUG(("read_block: Request sense: %s",
	      sense_msg[sensedata.extendedsense.SenseKey]));
	if(sensedata.extendedsense.Valid)
	  DEBUG(("read_block: in block %lu",
		sensedata.extendedsense.InfoBytes));
	switch(sensedata.extendedsense.SenseKey) {
	case 0:			/* NO SENSE */
	  DEBUG(("read_block: Got \"CHECK CONDITION\" - but no sense key !"));
	  goto OK;		/* The command seems to have executed
				   nice anyway - strange ! */
	  break;
	case 1:			/* RECOVERED ERROR */
	  goto OK;
	  break;
	case 2:			/* DEVICE NOT READY */
	  return EDRVNR;
	  break;
	case 3:			/* MEDIUM ERROR */
	  return EREADF;
	  break;
	case 4:			/* HARDWARE ERROR */
	  return EREADF;
	  break;
	case 5:			/* ILLEGAL REQUEST */
	  return EBADRQ;
	  break;
	case 6:			/* MEDIA CHANGED */
	  return E_CHNG;
	  break;
	case 7:			/* DATA PROTECT */
	  return EACCDN;
	  break;
	case 8:			/* BLANK CHECK */
	  return EWRITF;
	  break;
	case 9:			/* Device specific */
	  return ERROR;		/* Maybe return E_CHNG ? */
	  break;
	case 10:		/* COPY ABORTED */
	  return ERROR;
	  break;
	case 11:		/* ABORTED COMMAND, try again */
	  goto again;
	  break;
	case 12:		/* EQUAL - hmmm. Lets return error anyway */
	  return ERROR;
	  break;
	case 13:		/* VOLUME OVERFLOW */
	  return EWRITF;
	  break;
	case 14:		/* MISCOMPARE */
	  return ERROR;
	  break;
	case 15:		/* Reserved. */
	  return ERROR;
	  break;
	}
      }
      DEBUG(("Sensedata is not according to the SCSI standard !"));
    }
    return ERROR;		/* Probably a timeout */
  }
 OK:
  if(buf&&(nblocks==1)) {
    memcpy(buf,dmabuf,Options.block_size);
  }
  return 0;
}

int write_block(int device,long nblocks, char *buf) {
  int retcode;

  if(Options.Flags.test_unit_ready)
    if((retcode=test_unit_ready(device)))
      goto wjumphere;

wagain:
  TRACE(("write_block(%d, %ld, %lx)",device,nblocks,buf));
  if(nblocks==1) {
    retcode=scsi_seqwrite(device,1l,Options.block_size,
			  (buf)?buf:dmabuf,
			  (Options.timeouts[4]+Options.timeouts[3])*200l);
    DEBUG(("Returning from scsi_seqwrite"));
  } else {
    DEBUG(("write_block:  nblocks %ld, buf %06lx",
	   nblocks,buf));
    retcode=scsi_seqwrite(device,nblocks,Options.block_size,buf,
			  (Options.timeouts[4]*nblocks+Options.timeouts[3])*200l);
    DEBUG(("Returning from scsi_seqwrite"));
  }
 wjumphere:
  if(retcode) {
    DEBUG(("write_block: scsi_seqwrite returned %d",retcode));
    if((retcode&0x1e)==8) {
      DEBUG(("write_block: Device BUSY"));
      return EDRVNR;		/* Return Drive Not Ready */
    } else if((retcode&0x1e)==2) { /* OK, we will request sense... */
      SenseData sensedata;
      TRACE(("tapedev writeblock, requesting sense"));
      request_sense(device, &sensedata);
      DEBUG(("Valid: %d, FM: %d, EOM: %d, ILI: %d, ErrClass %d, ErrCode %d",
	     (sensedata.extendedsense.Valid)?1:0,
	     (sensedata.extendedsense.FM)?1:0,
	     (sensedata.extendedsense.EOM)?1:0,
	     (sensedata.extendedsense.ILI)?1:0,
	     sensedata.extendedsense.ErrorClass,
	     sensedata.extendedsense.ErrorCode));
      DEBUG(("SegmentNumber %d, InfoBytes %lx, ExtraBytes %d",
	     sensedata.extendedsense.SegmentNumber,
	     sensedata.extendedsense.InfoBytes,
	     sensedata.extendedsense.ExtraBytes));
      DEBUG(("CommandSpec: %lx, additionalSense %d, additionalQualifier %d",
	     sensedata.extendedsense.CommandSpec,
	     sensedata.extendedsense.additionalSense,
	     sensedata.extendedsense.additionalQualifier));
      DEBUG(("FieldReplacableUnitCode %d, SKSV %d, SenseSpecific %lx",
	     sensedata.extendedsense.FieldReplaceableUnitCode,
	     (sensedata.extendedsense.SKSV)?1:0,
	     sensedata.extendedsense.SenseSpecific));
      if(sensedata.nonextendedsense.ErrorClass<7) {
	ALERT(("write_block: Nonextended sense: class %u, code %u",
	      sensedata.nonextendedsense.ErrorClass,
	      sensedata.nonextendedsense.ErrorCode));
	if(sensedata.nonextendedsense.Valid)
	  ALERT(("write_block: in block %lu",
		((long)sensedata.nonextendedsense.BlockMSB<<16)|
		sensedata.nonextendedsense.BlockLSB));
	return EDRVNR;
      } else if(!sensedata.extendedsense.ErrorCode) {
	DEBUG(("write_block: Request sense: %d",
	       sensedata.extendedsense.SenseKey));
	if(sensedata.extendedsense.Valid)
	  DEBUG(("write_block: in block %lu",
		sensedata.extendedsense.InfoBytes));
	switch(sensedata.extendedsense.SenseKey) {
	case 0:			/* NO SENSE */
	  DEBUG(("write_block: Got \"CHECK CONDITION\" - but no sense key !"));
	  goto WOK;		/* The command seems to have executed
				   nice anyway - strange ! */
	  break;
	case 1:			/* RECOVERED ERROR */
	  goto WOK;
	  break;
	case 2:			/* DEVICE NOT READY */
	  return EDRVNR;
	  break;
	case 3:			/* MEDIUM ERROR */
	  return EREADF;
	  break;
	case 4:			/* HARDWARE ERROR */
	  return EREADF;
	  break;
	case 5:			/* ILLEGAL REQUEST */
	  return EBADRQ;
	  break;
	case 6:			/* MEDIA CHANGED */
	  return E_CHNG;
	  break;
	case 7:			/* DATA PROTECT */
	  return EWRPRO;
	  break;
	case 8:			/* BLANK CHECK */
	  return EWRITF;
	  break;
	case 9:			/* Device specific */
	  return ERROR;		/* Maybe return E_CHNG ? */
	  break;
	case 10:		/* COPY ABORTED */
	  return ERROR;
	  break;
	case 11:		/* ABORTED COMMAND, try again */
	  goto wagain;
	  break;
	case 12:		/* EQUAL - hmmm. Lets return error anyway */
	  return ERROR;
	  break;
	case 13:		/* VOLUME OVERFLOW */
	  return EWRITF;
	  break;
	case 14:		/* MISCOMPARE */
	  return ERROR;
	  break;
	case 15:		/* Reserved. */
	  return ERROR;
	  break;
	}
      }
      DEBUG(("Sensedata is not according to the SCSI standard !"));
    }
    return ERROR;		/* Probably a timeout */
  }
 WOK:
  if(nblocks==1)
    if(buf) 
      memcpy(dmabuf,buf,Options.block_size);
  return 0;
}

int do_skip(int device, FILEPTR *f, long amount) {
  long skippedthisfar;		/* Feature: If a file has been written */
  long nblocks;			/* to, and we did not fill the last buf */
  long error;			/* then this has not been flushed before */
  long i;			/* the seek !!! */
				/* Should not be a problem with TAR though */
  if(amount<0) {
    long offset=f->pos%Options.block_size;
    amount=-amount;
    if(amount<offset)
      return 0;
    amount-=offset;
    nblocks=(amount+Options.block_size-1)/Options.block_size;
    error=scsi_space(device, -nblocks, SPACE_BLOCKS,
		     nblocks*Options.timeouts[4]*200l+
		     Options.timeouts[2]*200l);
				/* If read_reverse had been more */
				/* generally supported, we could */
				/* have used that instead, because */
				/* then DMA would have been used */
    if(!error) {
      amount-=(nblocks*Options.block_size);
      if(amount<0)		/* Read in this block, so the buffer is */
	error=read_block(device,1l,NULL); /* up to date...  */
    }
    if(error) {
      return scsi2atarierr(device,error);
    }
  } else {
    skippedthisfar=min(amount,f->pos%Options.block_size);
    if(skippedthisfar==amount)
      return skippedthisfar;

    nblocks=(amount-skippedthisfar+Options.block_size-1)/Options.block_size;
#if 1
    error=scsi_space(device, nblocks, SPACE_BLOCKS,
		     nblocks*Options.timeouts[4]*200l+Options.timeouts[2]*200l);
    if(error)
      return scsi2atarierr(device,error);
#else
    for(i=0l;i<nblocks;i++) {
      error=read_block(device,1l,NULL);	/* Use read_block instead of SPACE */
      if(error)				/* because this uses DMA !!! */
	return scsi2atarierr(device,error);
    }
#endif
  }
  return 0;
}

char densitybuf[12];

long tape_open(FILEPTR *f) {
  long error;
  SenseHeader Senseheader;
  TRACE(("tape_open(%lx)",f));
  if(Options.Flags.mode_sense) {
    error=mode_sense(Options.device|(Options.lun<<8),0,0,sizeof(SenseHeader),&Senseheader);
    if(error) {
      if(error<0)
	return EDRVNR;
      return scsi2atarierr(Options.device|(Options.lun<<8),error);
    }
    Options.block_size=Senseheader.blockLength;
    DEBUG(("tape_open: blockLength %ld, numBlocks %ld, WriteProtect %d",
	   Senseheader.blockLength,Senseheader.numBlocks,Senseheader.WP));
  } 
  if(Options.Flags.mode_select) {
    bzero(densitybuf,512);
    densitybuf[2]=0x10;		/* Buffered mode, default speed */
    densitybuf[3]=8;
    densitybuf[4]=0;
    *(long *)(densitybuf+8)=Options.block_size;
    error=mode_select(Options.device|(Options.lun<<8),12,densitybuf);
    if(error) {
      if(error<0)
	return EDRVNR;
      return scsi2atarierr(Options.device|(Options.lun<<8),error);
    }
  }
  f->pos=0;
  f->devinfo=0l;
  return 0;
}

long tape_read(FILEPTR *f, char *buf, long bytes) {
  long error;
  long nblocks;
  long offset=f->pos%Options.block_size;
  long readchunk;
  long readsofar=0;

  DEBUG(("tape_read(FILE %lx,BUF %lx,BYTES %ld, offset %ld)",
	 f,buf,bytes,f->pos));

  if(bytes<=0)
    return 0l;

  if(offset) {			/* Fill in from last block read. */
    readchunk=min(Options.block_size-offset,bytes);
    memcpy(buf,(char *)(dmabuf+offset),readchunk);
    bytes-=readchunk;
    buf+=readchunk;
    readsofar=readchunk;
    f->pos+=readchunk;
  }
  nblocks=bytes/Options.block_size;
  if(nblocks) {
    error=read_block(Options.device|(Options.lun<<8),nblocks,buf);
    if(error) {
      return readsofar;
    }
    readchunk=nblocks*Options.block_size;
    buf=(char *)(buf+readchunk);
    bytes-=readchunk;
    readsofar+=readchunk;
    f->pos+=readchunk;
    nblocks=0;
  }
  if(bytes) {
    DEBUG(("tape_read : Reading %ld bytes from start of last block",
	   bytes));
    error=read_block(Options.device|(Options.lun<<8),1l,NULL);
    if(error) {
      return readsofar;
    }
    memcpy(dmabuf,buf,bytes);
    readsofar+=bytes;
    f->pos+=bytes;
    bytes=0;
  }
  return readsofar;
}

long tape_write(FILEPTR *f, const char *buf, long bytes) {
  long error;
  long nblocks;
  long offset=f->pos%Options.block_size;
  long writechunk;
  long writtensofar=0;

  DEBUG(("tape_write(FILE %lx,BUF %lx,BYTES %ld, offset %ld)",
	 f,buf,bytes,f->pos));

  if(bytes<=0)
    return 0l;

  f->devinfo=1;			/* Mark the file as being written to */
				/* so that we can flush it during close */

  if(offset) {			/* Fill in from last block read/write. */
    writechunk=min(Options.block_size-offset,bytes);
    memcpy((char *)(dmabuf+offset),buf,writechunk);
    bytes-=writechunk;
    buf+=writechunk;
    writtensofar=writechunk;
    f->pos+=writechunk;
    if(!(f->pos%Options.block_size)) {	/* We have filled out a block, now flush it */
      error=write_block(Options.device|(Options.lun<<8),1l,dmabuf);
      if(error) {
	return writtensofar;
      }
    }
  }
  nblocks=bytes/Options.block_size;
  if(nblocks) {
    error=write_block(Options.device|(Options.lun<<8),nblocks,buf);
    if(error) {
      return writtensofar;
    }
    writechunk=nblocks*Options.block_size;
    buf=(char *)(buf+writechunk);
    bytes-=writechunk;
    writtensofar+=writechunk;
    f->pos+=writechunk;
  }
  if(bytes) {
    DEBUG(("tape_write : Writing %ld bytes from start of last block",
	   bytes));
    memcpy(dmabuf,buf,bytes);
    writtensofar+=bytes;
    f->pos+=bytes;
  }
  return writtensofar;
}

long tape_lseek(FILEPTR *f, long where, int whence) {
  long error;
  TRACE(("tape_lseek(%lx,%ld,%d)",f,where,whence));
  switch(whence) {
  case SEEK_SET:
    if(where<0)
      return ERANGE;

    if((error=do_skip(Options.device|(Options.lun<<8),f,where-f->pos))<0) {
      DEBUG(("tape_lseek: do_skip returned %ld"));
      return error;
    }
    f->pos=where;
    break;
  case SEEK_CUR:
    if(where+f->pos<0)
      return ERANGE;
    if((error=do_skip(Options.device|(Options.lun<<8),f,where))<0) {
      DEBUG(("tape_lseek: do_skip returned %ld"));
      return error;
    }
    f->pos+=where;
    break;
  case SEEK_END:
    return EINVFN;
    break;
  default:
    return EINVFN;
  }
  return f->pos;
}

long tape_mount(void) {
  long error=load_unload(Options.device|(Options.lun<<8),Load,
			 Options.Flags.immed,0,0,Options.timeouts[0]*200l);
  if(error)
    return scsi2atarierr(Options.device|(Options.lun<<8),error);
  if(Options.Flags.lock) {
    error=prevent_allow_removal(Options.device|(Options.lun<<8),1);
    if(error)
      return scsi2atarierr(Options.device|(Options.lun<<8),error);
  }
  return 0l;
}

long tape_unmount(void) {
  long error=load_unload(Options.device|(Options.lun<<8),Unload,
			 Options.Flags.immed,0,0,Options.timeouts[0]*200l);
  if(error)
    return scsi2atarierr(Options.device|(Options.lun<<8),error);
  if(Options.Flags.lock) {
    error=prevent_allow_removal(Options.device|(Options.lun<<8),0);
    if(error)
      return scsi2atarierr(Options.device|(Options.lun<<8),error);
  }
  return 0l;
}

long tape_rewind(void) {
  long error=scsi_rewind(Options.device|(Options.lun<<8),Options.Flags.immed,Options.timeouts[1]*200l);
  if(error)
    return scsi2atarierr(Options.device|(Options.lun<<8),error);
  return 0l;
}

long tape_write_filemark(long nmarks) {
  long error=scsi_filemark(Options.device|(Options.lun<<8),nmarks,
			   (Options.timeouts[3]+nmarks*Options.timeouts[4])*200l);
  if(error)
    return scsi2atarierr(Options.device|(Options.lun<<8),error);
  return 0l;
}

long tape_skip_filemark(long nmarks) {
  long error=scsi_space(Options.device|(Options.lun<<8),nmarks,SPACE_FILEMARKS,
			Options.timeouts[1]*200);
  if(error)
    return scsi2atarierr(Options.device|(Options.lun<<8),error);
  return 0l;
}

long tape_ioctl(FILEPTR *f, int mode, void *buf) {
  TRACE(("tape_ioctl(%lx,%d,%lx)",f,mode,buf));
  switch(mode) {
  case FIONREAD:
    *((long *)buf)=256l*Options.block_size;
    return 0;
  case FIONWRITE:
    *((long *)buf)=256l*Options.block_size;
    return 0;
  case TMAGIC:
    *((long *)buf)=TAPEMAGIC;
    return 0;
    break;
  case TMOUNT:
    return tape_mount();
    break;
  case TUMOUNT:
    return tape_unmount();
    break;
  case TREW:
    return tape_rewind();
    break;
  case TWFILEM:
    return tape_write_filemark(*(long *)buf);
    break;
  case TSFILEM:
    return tape_skip_filemark(*(long *)buf);
    break;
  case TOPTIONS:
    memcpy(&Options, (struct options *)buf,sizeof(struct options));
    return 0l;
  default:
    return EINVFN;
  }
}

long tape_datime(FILEPTR *f, short *timeptr, int rwflag) {
  TRACE(("tape_datime(%lx,%lx,%d)",f,timeptr,rwflag));
  if(rwflag) {
    inittime=timeptr[0];
    initdate=timeptr[1];
  } else {
    timeptr[0]=inittime;
    timeptr[1]=initdate;
  }
  return 0l;
}

long tape_close(FILEPTR *f, int pid) {
  long error;
  DEBUG(("Last position was %ld",f->pos));
  if(f->devinfo) {
    if(f->pos%Options.block_size) {	/* Flush buffer !!! */
      error=write_block(Options.device|(Options.lun<<8), 1l, NULL);
      if(error) {
	if(error<0)
	  return EDRVNR;
	return scsi2atarierr(Options.device|(Options.lun<<8),error);
      }
    }
    error=scsi_filemark(Options.device|(Options.lun<<8),1l,30000l);	/* Mark End-of-data */
    if(error) {
      if(error<0)
	return EDRVNR;
      return scsi2atarierr(Options.device|(Options.lun<<8),error);
    }
  }
  return 0l;
}

long tape_select(FILEPTR *f, long proc, int mode) {
  TRACE(("tape_select()"));
  switch(mode) {
  case O_WRONLY:
    return 1;
    break;	
  case O_RDONLY:
    return 1;
    break;
  }
  return 1;
}

void tape_unselect(FILEPTR *f, long proc, int mode) {
  TRACE(("tape_unselect()"));
  return;
}

#ifndef PROG_DEVICE
DEVDRV *tape_init(struct kerinfo *k) {
  kernel=k;
  CCONWS("\033pTAPE device driver\033q\r\nVersion ");
  CCONWS(version);
  CCONWS("\r\n(c) Erling Henanger 1994\r\n");
  load_config();
  if(DCNTL(DEV_INSTALL,Options.mountpoint,&devinfo)<=0l) {
    CCONWS("Unable to install tape device\r\n");
    return (DEVDRV *)0l;
  }
  inittime=TGETTIME();
  initdate=TGETDATE();
  CCONWS("Tape device installed\r\n");
  return (DEVDRV *)1l;
}
#else
int main() {
  Cconws("\033pTAPE device driver\033q\r\nVersion ");
  Cconws(version);
  Cconws("\r\n(c) Erling Henanger 1994\r\n");
  load_config();
  if((long)(kernel=(struct kerinfo *)Dcntl(DEV_INSTALL,
				     Options.mountpoint,&devinfo))<=0l) {
    Cconws("Unable to install tape device\r\n");
    return 1;
  }
  inittime=Tgettime();
  initdate=Tgetdate();
  Cconws("Tape device installed\r\n");
  Ptermres(256L + _base->p_tlen + _base->p_dlen + _base->p_blen, 0);
}
#endif
